#!/usr/bin/env python3

"""subprocess_commands.py"""

# pylint: disable=C0116 missing-function-docstring

import os
import time
import subprocess
import signal

from roboauto.global_state import roboauto_state, roboauto_options
from roboauto.logger import print_out, print_err
from roboauto.utils import file_is_executable, json_dumps


def subprocess_run_command(program, error_print=True):
    try:
        process = subprocess.run(program, capture_output=True, check=False)
    except FileNotFoundError:
        program_name = program[0]
        print_err(f"command {program_name} does not exists")
        return False
    if process.returncode != 0:
        if error_print:
            print_err(process.stderr.decode(), end="", error=False, date=False)
        return False

    return process.stdout


def message_notification_send(event, message):
    if not file_is_executable(roboauto_state["message_notification_command"]):
        print_err("message notification command not found, no messages will be sent")
        return False

    message_stdout = subprocess_run_command([
        roboauto_state["message_notification_command"],
        event, message
    ])
    if message_stdout is False:
        print_err(f"sending message {event}")
        return False

    return message_stdout.decode()


def subprocess_generate_invoice(invoice_amout, invoice_label):
    invoice_generate_output = subprocess_run_command([
        roboauto_state["lightning_node_command"], "invoice",
        invoice_amout, invoice_label
    ])
    if invoice_generate_output is False:
        print_err("generating the invoice")
        return False

    return invoice_generate_output.decode()


def subprocess_pay_invoice_and_check(
    robot_dic, order_id,
    invoice, amount_satoshis_string,
    pay_label, is_paid_function,
    string_checking, string_paid, string_not_paid,
    order_dic_function, failure_function,
    maximum_retries=None
):
    robot_name = robot_dic["name"]

    check_output = subprocess_run_command([
        roboauto_state["lightning_node_command"], "check",
        invoice, amount_satoshis_string
    ])
    if check_output is False:
        print_err(
            "lightning-node check returned false, invoice will not be paid"
        )
        return False
    print_out(check_output.decode(), end="", date=False)
    print_out("invoice checked successfully")

    pay_command = [
        roboauto_state["lightning_node_command"], "pay",
        invoice, pay_label
    ]

    retries = 0
    with subprocess.Popen(pay_command, start_new_session=True) as pay_subprocess:
        while True:
            if maximum_retries is not None and retries > maximum_retries:
                print_err("maximum retries occured for pay command")
                return False

            print_out(string_checking)

            order_dic = order_dic_function(robot_dic, order_id)
            if order_dic is False or order_dic is None or isinstance(order_dic, str):
                return False

            order_response_json = order_dic["order_response_json"]

            order_status = order_response_json.get("status", False)
            if order_status is False:
                print_err(json_dumps(order_response_json), error=False, date=False)
                print_err(f"getting order_status of {robot_name} {order_id}")
                return False

            if is_paid_function(order_status):
                if not failure_function(order_status):
                    print_out(robot_name + " " + string_paid)
                    return_output = order_dic
                else:
                    print_err(robot_name + " " + string_not_paid)
                    return_output = False
                try:
                    os.killpg(os.getpgid(pay_subprocess.pid), signal.SIGTERM)
                except ProcessLookupError:
                    pass

                return return_output

            retries += 1
            time.sleep(roboauto_options["pay_interval"])



#!/usr/bin/env python3

"""logger.py"""

# pylint: disable=C0116 missing-function-docstring

import sys

from roboauto.global_state import roboauto_state
from roboauto.date_utils import date_get_current


def logger_flush():
    if roboauto_state["logger"] is not None:
        roboauto_state["logger"].flush()


def print_and_log(arg, end="\n", file_stream=None, level: int | bool = True):
    """level is for writing to stdout/stderr, everything is always
    saved to the log file.
    level goes from 0 (True) -> always log also to the terminal
    to higher integers that will be logged to the terminal just
    when roboauto_state["log_level"] is high enough"""

    if level is True:
        level_int = 0
    else:
        level_int = level

    if level_int <= roboauto_state["log_level"]:
        if file_stream is None:
            print(arg, end=end)
        else:
            print(arg, end=end, file=file_stream)

    if roboauto_state["current_command_type"] != "info":
        if roboauto_state["logger"] is None:
            log_file_name = \
                roboauto_state["log_home"] + "/" + \
                roboauto_state["current_command_type"] + \
                ".log"

            try:
                # pylint: disable=R1732 consider-using-with
                roboauto_state["logger"] = open(
                    log_file_name, "a", encoding="utf8"
                )
            except OSError:
                return False

        print(arg, end=end, file=roboauto_state["logger"])

    return True


def print_out(arg, end="\n", date=True, level: int | bool = True):
    string_print = arg

    if roboauto_state["current_command_type"] == "keep-online" and date:
        date_current = date_get_current()
        string_print = "[" + date_current + "] " + string_print

    print_and_log(string_print, end=end, level=level)


def print_stderr(
    error_string, arg, end="\n",
    date=True, error=True, level: int | bool = True
):
    string_print = arg

    if error:
        string_print = error_string + string_print

    if roboauto_state["current_command_type"] == "keep-online" and date:
        date_current = date_get_current()
        string_print = "[" + date_current + "] " + string_print

    print_and_log(string_print, end=end, file_stream=sys.stderr, level=level)


def print_war(arg, end="\n", date=True, error=True, level: int | bool = True):
    print_stderr("warning: ", arg, end=end, date=date, error=error, level=level)


def print_err(arg, end="\n", date=True, error=True, level: int | bool = True):
    print_stderr("error: ", arg, end=end, date=date, error=error, level=level)



#!/usr/bin/env python3

"""gpg_key.py"""

# pylint: disable=C0116 missing-function-docstring

import gnupg

from roboauto.global_state import roboauto_state
from roboauto.logger import print_err
from roboauto.date_utils import date_get_yesterday
from roboauto.utils import token_get_double_sha256


def gpg_get():
    if roboauto_state["gpg"] is None:
        roboauto_state["gpg"] = \
            gnupg.GPG(gnupghome=roboauto_state["gnupg_home"], verbose=False)

    return roboauto_state["gpg"]


def gpg_generate_robot(token):
    gpg = gpg_get()

    token_double_sha256 = token_get_double_sha256(token)

    key_passphrase = token

    yesterday = date_get_yesterday(date_format="%Y-%m-%d")

    # stupid python-gnupg put by default Name-Email in gen_key_input and no option to remove it
    input_data = f"""\
Key-Type: EDDSA
Key-Curve: ed25519
Key-Usage: sign
Subkey-Type: ECC
Subkey-Curve: cv25519
SubKey-Usage: encrypt
Passphrase: {key_passphrase}
Name-Real: RoboSats ID {token_double_sha256}
Creation-Date: {yesterday}
%commit
"""

    key = gpg.gen_key(input_data)

    fingerprint = key.fingerprint

    public_key = gpg.export_keys(fingerprint, secret=False, armor=True)
    private_key = gpg.export_keys(
        fingerprint, secret=True, armor=True,
        passphrase=key_passphrase
    )

    return fingerprint, public_key, private_key


def gpg_import_key(key, set_trust=True, passphrase=None, error_print=True):
    gpg = gpg_get()

    key = gpg.import_keys(key, passphrase=passphrase)

    if not hasattr(key, "count") or key.count != 1:
        if error_print:
            print_err("imported key is multiple keys")
        return False

    fingerprint = key.fingerprints[0]

    if set_trust is True:
        gpg.trust_keys(fingerprint, "TRUST_ULTIMATE")

    return fingerprint


def gpg_encrypt_sign_message(
    message, fingerprints, sender_fingerprint, passphrase=None, error_print=True
):
    gpg = gpg_get()

    # also set yesterday date (maybe ?) (how ?) as in robosats/frontend/src/pgp/index.ts
    encrypted_data = gpg.encrypt(
        message, fingerprints, sign=sender_fingerprint, passphrase=passphrase
    )

    if encrypted_data.ok is not True:
        if error_print:
            print_err("encrypting and signing message")
        return False

    message_enc = str(encrypted_data)

    return message_enc


def gpg_decrypt_check_message(
    message, fingerprint_signature, passphrase=None, error_print=True
):
    gpg = gpg_get()

    decrypted_data = gpg.decrypt(message, passphrase=passphrase)

    if decrypted_data.ok is not True:
        if error_print:
            print_err("decrypting and checking message")
        return False, False

    decrypted_message = str(decrypted_data)

    signature_error = None

    if not hasattr(decrypted_data, "fingerprint"):
        signature_error = "message is not signed"

    if decrypted_data.fingerprint != fingerprint_signature:
        signature_error = "message is not signed with the correct key"

    if signature_error is not None and error_print:
        print_err(signature_error)

    return decrypted_message, signature_error


def gpg_sign_message(message, fingerprint, passphrase=None, error_print=True):
    gpg = gpg_get()

    signed_data = gpg.sign(
        message, keyid=fingerprint, passphrase=passphrase,
        extra_args=["--digest-algo", "SHA512"]
    )

    if hasattr(signed_data, "status_detail") and signed_data.status_detail is not None:
        if error_print:
            print_err("signing message")
        return False

    signed_message = str(signed_data)

    return signed_message



#!/usr/bin/env python3

"""order_local.py"""

# pylint: disable=C0116 missing-function-docstring
# pylint: disable=C0209 consider-using-f-string

import os
import re
import shutil

from roboauto.logger import print_out, print_err
from roboauto.global_state import roboauto_state, roboauto_options
from roboauto.date_utils import \
    date_convert_time_zone_and_format_string, \
    get_current_timestamp, get_hour_offer, get_current_hour_from_timestamp
from roboauto.utils import \
    json_dumps, file_json_read, is_float, get_int, file_remove, \
    dir_make_sure_exists, file_json_write, directory_get_file_numbers
from roboauto.subprocess_commands import message_notification_send
from roboauto.order_data import \
    get_currency_string, order_is_pending, get_type_string, \
    order_is_public, order_is_paused, order_is_waiting_taker_bond
from roboauto.robot import \
    robot_list_dir, robot_get_dir_dic, \
    robot_load_from_name


def get_order_data(
    type_id, currency_id,
    amount, has_range, min_amount, max_amount,
    payment_method, premium,
    public_duration, escrow_duration, bond_size
):
    return {
        "type":                 type_id,
        "currency":             currency_id,
        "amount":               amount,
        "has_range":            has_range,
        "min_amount":           min_amount,
        "max_amount":           max_amount,
        "payment_method":       payment_method,
        "premium":              premium,
        "public_duration":      public_duration,
        "escrow_duration":      escrow_duration,
        "bond_size":            bond_size
    }


def get_order_user(
    type_string, currency_string, min_amount_user, max_amount_user,
    payment_method, premium, public_duration, escrow_duration, bond_size
):
    return {
        "type":                 type_string,
        "currency":             currency_string,
        "min_amount":           min_amount_user,
        "max_amount":           max_amount_user,
        "payment_method":       payment_method,
        "premium":              premium,
        "public_duration":      public_duration,
        "escrow_duration":      escrow_duration,
        "bond_size":            bond_size
    }


def get_offer_dic(offer, coordinator):
    offer_id = offer.get("id", "")
    expires_at = offer.get("expires_at", "")
    order_type_bool = offer.get("type", "")
    currency_id = offer.get("currency", "")
    amount = offer.get("amount", "")
    has_range = offer.get("has_range", "")
    min_amount = offer.get("min_amount", "")
    max_amount = offer.get("max_amount", "")
    payment_method = offer.get("payment_method", "")
    premium = offer.get("premium", "")
    escrow_duration_seconds = offer.get("escrow_duration", "")
    bond_size = offer.get("bond_size", "")
    maker_nick = offer.get("maker_nick", "")
    maker_status = offer.get("maker_status", "")

    if order_type_bool == 0:
        order_type = "buy"
    elif order_type_bool == 1:
        order_type = "sell"
    elif order_type_bool == 2:
        order_type = "all"
    else:
        order_type = "???"

    date_end = date_convert_time_zone_and_format_string(
        expires_at,
        output_format="%H:%M:%S"
    )

    currency = str(get_currency_string(currency_id)).lower()

    if not has_range:
        min_amount = amount
        max_amount = amount

    if currency != "btc":
        amount_format = "%7.0f"
    else:
        amount_format = "%7.3f"

    duration = str(int(escrow_duration_seconds / 3600))
    if maker_nick in robot_list_dir(roboauto_state["active_home"]):
        ours = " * "
    else:
        ours = " - "

    if maker_status == "Seen recently":
        maker_status = "Recently"

    offer_dic = {
        "coordinator": str(coordinator),
        "offer_id": offer_id,
        "maker_nick": maker_nick,
        "maker_status": maker_status,
        "order_type": order_type,
        "currency": currency,
        "duration": duration,
        "bond_size": bond_size,
        "premium": premium,
        "ours": ours,
        "min_amount": min_amount,
        "max_amount": max_amount,
        "date_end": date_end,
        "payment_method": payment_method,
        "amount_format": amount_format
    }

    return offer_dic


def offer_dic_print(offer_dic):
    # when changing change also robot_order_not_complete_print
    printf_string = \
        "%-3s %-6s %-8s %-24s %-4s %-3s %3sh %5s %6.2f%% %3s " + \
        offer_dic["amount_format"] + " " + offer_dic["amount_format"] + \
        " %8s %s"
    print_out(printf_string % (
        offer_dic["coordinator"][:3],
        offer_dic["offer_id"],
        offer_dic["maker_status"], offer_dic["maker_nick"],
        offer_dic["order_type"], offer_dic["currency"],
        offer_dic["duration"], offer_dic["bond_size"], float(offer_dic["premium"]),
        offer_dic["ours"],
        float(offer_dic["min_amount"]), float(offer_dic["max_amount"]),
        offer_dic["date_end"],
        offer_dic["payment_method"]
    ))


def robot_have_orders_dir(robot_dir):
    if os.path.isdir(robot_dir + "/orders"):
        return True

    return False


def robot_have_make_response(robot_dir):
    if os.path.isfile(robot_dir + "/make-response"):
        return True

    return False


def robot_have_make_data(robot_dir):
    if os.path.isfile(robot_dir + "/make-data"):
        return True

    return False


def robot_set_make_response(robot_dir, make_response_json):
    if not file_json_write(robot_dir + "/make-response", make_response_json):
        return False

    return True


def robot_get_make_response(robot_dir, error_print=True):
    make_response_file = robot_dir + "/make-response"

    if not os.path.isfile(make_response_file):
        return False

    make_response_json = file_json_read(make_response_file, error_print=error_print)
    if make_response_json is False:
        return False

    return make_response_json


def robot_order_set_local_make_data(robot_dir, make_data):
    if file_json_write(robot_dir + "/make-data", make_data) is False:
        return False

    return True


def robot_order_get_local_make_data(robot_dir, error_print=True):
    # remember to remove the make data after creating the order
    local_make_data_file = robot_dir + "/make-data"

    if not os.path.isfile(local_make_data_file):
        return False

    make_data = file_json_read(local_make_data_file, error_print=error_print)
    if make_data is False:
        return False

    return make_data


def robot_order_remove_local_make_data(robot_dir):
    local_make_data_file = robot_dir + "/make-data"

    if not os.path.isfile(local_make_data_file):
        return True

    if not file_remove(local_make_data_file):
        return False

    return True


def order_id_list_from_robot_dir(robot_dir, error_print=True):
    orders_dir = robot_dir + "/orders"
    if not os.path.isdir(orders_dir):
        if error_print:
            print_err(f"{orders_dir} is not a dir")
        return None

    order_ids = directory_get_file_numbers(orders_dir)
    if order_ids == 0:
        if error_print:
            print_err(f"orders dir {orders_dir} is empty")
        return False
    elif order_ids is False:
        return False

    return order_ids


def order_id_last_from_robot_dir(robot_dir, error_print=True):
    order_ids = order_id_list_from_robot_dir(robot_dir, error_print=error_print)
    if order_ids is False or order_ids is None:
        return False

    return str(order_ids[-1])


def order_robot_get_last_order_id(robot_dic, error_print=True):
    """return a string"""
    robot_dir = robot_dic["dir"]

    maximum_order_id = False

    order_ids = order_id_list_from_robot_dir(robot_dir, error_print=error_print)
    if order_ids is not False and order_ids is not None:
        maximum_order_id = order_ids[-1]

    make_data = robot_get_make_response(robot_dir, error_print=error_print)
    if make_data is not False and make_data is not None:
        make_id = make_data.get("id", False)
        if make_id is not False and make_id > maximum_order_id:
            maximum_order_id = make_id

    if isinstance(maximum_order_id, int):
        return str(maximum_order_id)
    else:
        return False


def order_dic_from_robot_dir(robot_dir, order_id=None, error_print=True):
    orders_dir = robot_dir + "/orders"
    if not os.path.isdir(orders_dir):
        if error_print:
            print_err(f"{orders_dir} is not a dir")
        return None

    if order_id is False or order_id is None:
        order_id = order_id_last_from_robot_dir(robot_dir, error_print=error_print)
        if order_id is False or order_id is None:
            return False

    order_file = orders_dir + "/" + order_id
    if not os.path.isfile(order_file):
        if error_print:
            print_err(f"orders dir {orders_dir} does not have order {order_id}")
        return False

    order_dic = file_json_read(order_file)
    if order_dic is False:
        return False

    return order_dic


def robot_order_not_complete_print(robot_name, coordinator, error_string):
    print_out(
        "%-3s %-6s %-8s %-24s %4s %3s %4s %5s %7s %3s %7s %7s %8s %s" % (
            str(coordinator)[:3], "------", "--------", robot_name,
            "----", "---", "----", "-----", "-------",
            "---", "-------", "-------", "--------",
            error_string
        )
    )


def robot_no_order_dir_print(robot_name, coordinator):
    robot_order_not_complete_print(robot_name, coordinator, "no order dir")


def robot_no_order_response_print(robot_name, coordinator):
    robot_order_not_complete_print(robot_name, coordinator, "no order response")


def order_dic_print(order_dic, robot_name, coordinator, one_line, full_mode):
    # pylint: disable=R1702 too-many-nested-blocks

    if not one_line:
        if full_mode:
            print_out(json_dumps(order_dic))
        else:
            if "order_user" not in order_dic and "order_info" not in order_dic:
                print_out(json_dumps({"error": "no order user and info"}))
            else:
                order_dic_list = {}

                if "order_info" in order_dic:
                    for key in (
                        "coordinator", "order_id", "status_string",
                        "peer_nick", "amount_single"
                    ):
                        if key in order_dic["order_info"]:
                            order_dic_list.update({key: order_dic["order_info"][key]})

                if "order_user" in order_dic:
                    order_dic_list.update(order_dic["order_user"])

                print_out(json_dumps(order_dic_list))
    else:
        if "order_response_json" in order_dic:
            offer_dic_print(get_offer_dic(
                order_dic["order_response_json"], coordinator
            ))
        else:
            robot_no_order_response_print(robot_name, coordinator)


def order_info_local_print_ordered_list(robot_list):
    order_list_unsorted = []
    robot_list_no_response = []
    robot_list_no_dir = []
    for robot_name in robot_list:
        robot_dic = robot_load_from_name(robot_name, error_print=False)
        if robot_dic is False:
            continue

        order_dic = order_dic_from_robot_dir(
            robot_dic["dir"], order_id=None, error_print=False
        )
        if order_dic is None:
            robot_list_no_dir.append(robot_dic)
        elif order_dic is not False:
            if order_dic.get("order_response_json", False) is not False:
                order_list_unsorted.append({
                    "order_dic": order_dic,
                    "robot_name": robot_name,
                    "coordinator": robot_dic["coordinator"]
                })
            else:
                robot_list_no_response.append(robot_dic)

    for robot_dic in robot_list_no_dir:
        robot_no_order_dir_print(robot_dic["name"], robot_dic["coordinator"])
    for robot_dic in robot_list_no_response:
        robot_no_order_response_print(robot_dic["name"], robot_dic["coordinator"])

    order_list_sorted = sorted(
        order_list_unsorted,
        key=lambda order_data: float(order_data["order_dic"]["order_response_json"]["premium"])
    )
    for order_data in order_list_sorted:
        order_dic_print(
            order_data["order_dic"], order_data["robot_name"], order_data["coordinator"],
            one_line=True, full_mode=False
        )

    return True


def order_info_dir(argv):
    if len(argv) < 1:
        print_err("insert arguments")
        return False

    first_arg = argv[0]
    argv = argv[1:]
    if first_arg in ("--active", "--pending", "--paused", "--inactive"):
        destination_dir = robot_get_dir_dic()[first_arg[2:]]

        if order_info_local_print_ordered_list(
            os.listdir(destination_dir)
        ) is False:
            return False
    elif first_arg == "--dir":
        if len(argv) < 1:
            print_err("insert directory")
            return False
        robot_dir = argv[0]
        argv = argv[1:]

        if not os.path.isdir(robot_dir):
            print_err(f"{robot_dir} is not a directory")
            return False
        if order_info_local_print_ordered_list(
            os.listdir(robot_dir)
        ) is False:
            return False
    elif re.match('^-', first_arg) is not None:
        print_err(f"option {first_arg} not recognized")
        return False
    else:
        print_err(f"argument {first_arg} not recognized")
        return False

    return True


def order_summary(argv):
    if len(argv) < 1:
        print_err("insert arguments")
        return False

    first_arg = argv[0]
    argv = argv[1:]

    summary_dic = {
        "buy": {},
        "sell": {}
    }
    coordinator_number = {}
    for federation_coordinator in roboauto_options["federation"]:
        coordinator_number[federation_coordinator] = 0

    if first_arg in ("--active", "--pending", "--paused", "--inactive"):
        for robot_name in os.listdir(robot_get_dir_dic()[first_arg[2:]]):
            robot_dic = robot_load_from_name(robot_name, error_print=False)
            if robot_dic is False:
                continue

            robot_dir = robot_dic["dir"]
            coordinator = robot_dic["coordinator"]

            order_dic = order_dic_from_robot_dir(
                robot_dir, order_id=None, error_print=False
            )
            if order_dic is not None and order_dic is not False:
                order_user = order_dic.get("order_user", False)
                if order_user is False:
                    continue
                type_string = order_user["type"]
                currency_string = order_user["currency"]
            else:
                make_data = robot_get_make_response(robot_dir)
                if make_data is False:
                    make_data = robot_order_get_local_make_data(robot_dir)
                    if make_data is False:
                        continue
                type_string = str(get_type_string(
                    make_data.get("type", False)
                ))
                currency_string = str(get_currency_string(
                    make_data.get("currency", False)
                )).lower()

            if currency_string not in summary_dic[type_string]:
                summary_dic[type_string][currency_string] = {}
                for federation_coordinator in roboauto_options["federation"]:
                    summary_dic[type_string][currency_string][federation_coordinator] = 0

            if coordinator not in summary_dic[type_string][currency_string]:
                summary_dic[type_string][currency_string][coordinator] = 0
                coordinator_number[coordinator] = 0

            summary_dic[type_string][currency_string][coordinator] += 1
            coordinator_number[coordinator] += 1

        summary_sorted = {}
        for key, value in summary_dic.items():
            summary_sorted[key] = dict(sorted(value.items()))
            for currency in summary_sorted[key]:
                total_currency = 0
                for coordinator in summary_sorted[key][currency]:
                    total_currency += summary_sorted[key][currency][coordinator]
                summary_sorted[key][currency]["TOT"] = total_currency

        coordinator_total = 0
        for _, number in coordinator_number.items():
            coordinator_total += number
        coordinator_number["TOT"] = coordinator_total

        print_out(json_dumps(summary_sorted))
        print_out(json_dumps(coordinator_number))

    return True


def order_save_order_file(robot_dir, order_id, order_dic):
    orders_dir = robot_dir + "/orders"
    if not dir_make_sure_exists(orders_dir):
        return False
    order_file = orders_dir + "/" + order_id
    if not file_json_write(order_file, order_dic):
        print_err("saving order %s to file" % order_id)
        return False

    return True


def robot_handle_taken(robot_name, status_id, order_id, other):
    robot_dir = roboauto_state["active_home"] + "/" + robot_name
    if not os.path.isdir(robot_dir):
        print_err(robot_name + " is not active")
        return False

    if order_is_pending(status_id):
        dest_dir = roboauto_state["pending_home"]
        dest_name = "pending"
    else:
        dest_dir = roboauto_state["inactive_home"]
        dest_name = "inactive"

    print_out(f"{robot_name} {order_id} something happened")
    print_out(f"{robot_name} moved to {dest_name}")

    try:
        shutil.move(robot_dir, dest_dir)
    except OSError:
        print_err(
            "moving %s to %s" %
            (robot_dir, dest_dir)
        )
        return False

    message_output = message_notification_send(
        "order-taken",
        robot_name + " " + order_id + " " + other
    )
    if message_output is False:
        return False

    print_out(message_output, end="", date=False)

    return True


def order_data_from_order_user(order_user):
    type_string = order_user.get("type", False)
    if not type_string:
        print_err("type not present")
        return False

    currency_string = order_user.get("currency", False)
    if not currency_string:
        print_err("currency not present")
        return False

    min_amount = order_user.get("min_amount", False)
    if not min_amount:
        print_err("min_amount not present")
        return False

    max_amount = order_user.get("max_amount", False)
    if not max_amount:
        print_err("max_amount not present")
        return False

    payment_method = order_user.get("payment_method", False)
    if not payment_method:
        print_err("payment_method not present")
        return False

    premium = order_user.get("premium", False)
    if not premium:
        print_err("premium not present")
        return False

    public_duration_str = order_user.get("public_duration", False)
    if not public_duration_str:
        print_err("public_duration not present")
        return False

    escrow_duration_str = order_user.get("escrow_duration", False)
    if not escrow_duration_str:
        print_err("escrow_duration not present")
        return False

    bond_size = order_user.get("bond_size", False)
    if not bond_size:
        print_err("bond_size not present")
        return False

    type_id = int(get_type_string(type_string, reverse=True))
    if type_id < 0:
        print_err("type %s is not valid" % type_string)
        return False

    currency_id = int(get_currency_string(currency_string, reverse=True))
    if currency_id < 0:
        print_err("currency %s is not valid" % currency_string)
        return False

    if not is_float(min_amount, additional_check="positive"):
        print_err("min amount %s is not a positive number" % min_amount)
        return False

    if not is_float(max_amount, additional_check="positive"):
        print_err("max amount %s is not a positive number" % max_amount)
        return False

    if min_amount != max_amount:
        has_range = True
        amount = None
    else:
        has_range = False
        amount = min_amount

    if len(payment_method) > 70:
        print_err("payment method is longer than 70 characters")
        return False

    if not is_float(premium):
        print_err("premium %s is not a number" % premium)
        return False

    public_duration = get_int(public_duration_str)
    if public_duration is False:
        return False
    min_public_duration = 600
    max_public_duration = 86400
    if public_duration < min_public_duration or public_duration > max_public_duration:
        print_err("public duration should be between %d and %d" % (
            min_public_duration, max_public_duration
        ))
        return False

    escrow_duration = get_int(escrow_duration_str)
    if escrow_duration is False:
        return False
    min_escrow_duration = 1800
    max_escrow_duration = 28800
    if escrow_duration < min_escrow_duration or escrow_duration > max_escrow_duration:
        print_err("escrow duration should be between %d and %d" % (
            min_escrow_duration, max_escrow_duration
        ))
        return False

    if not is_float(bond_size, additional_check="percentage"):
        print_err("bond size %s is not a number between 0 and 100" % bond_size)
        return False

    return get_order_data(
        type_id, currency_id,
        amount, has_range, min_amount, max_amount,
        payment_method, premium,
        public_duration, escrow_duration, bond_size
    )


def order_is_this_hour_and_online(order, coordinator=False):
    current_timestamp = get_current_timestamp()

    order_info = order.get("order_info", False)
    if order_info is False:
        return False

    if not isinstance(coordinator, bool):
        order_coordinator = order_info.get("coordinator", False)
        if order_coordinator is False:
            return False
        if coordinator[:3] != order_coordinator[:3]:
            return False

    order_response_json = order.get("order_response_json", False)
    if order_response_json is False:
        return False

    expires_at = order_response_json.get("expires_at", False)
    if expires_at is False:
        return False

    date_hour = get_hour_offer(
        expires_at, current_timestamp,
        roboauto_state["keep_online_hour_relative"]
    )
    if date_hour is False:
        return False
    if date_hour != get_current_hour_from_timestamp(current_timestamp):
        return False

    status_id = order_info.get("status", False)
    if status_id is False:
        return False

    if \
        not order_is_public(status_id) and \
        not order_is_paused(status_id) and \
        not order_is_waiting_taker_bond(status_id):
        return False

    return True



#!/usr/bin/env python3

"""utils.py"""

# pylint: disable=C0116 missing-function-docstring
# pylint: disable=C0209 consider-using-f-string

import os
import re
import getpass
import json
import configparser
import random
import secrets
import hashlib
import struct

from roboauto.logger import print_out, print_err
from roboauto.global_state import roboauto_options, roboauto_state


def roboauto_first_coordinator():
    return next(iter(roboauto_options["federation"]))


# if this is called with robot_dic["coordinator"] do not check for False
# already checked when creating the robot_dic in robot_get_dic
def roboauto_get_coordinator_url(coordinator):
    url = roboauto_options["federation"].get(coordinator, False)
    if url is False:
        print_err(f"coordinator {coordinator} not valid")
        return False

    return url


def roboauto_get_coordinator_from_url(coordinator_url):
    for coordinator in roboauto_options["federation"]:
        if roboauto_options["federation"][coordinator] == coordinator_url:
            return coordinator

    return "---"


def get_coordinator_from_param(param):
    multi_false = False, False

    coordinator_value = param[2:]
    if len(coordinator_value) < 3:
        print_err(
            "coordinator name should be at least 3 characters long: "
            f"{coordinator_value} invalid"
        )
        return multi_false
    coordinator_found = False
    coordinator = ""
    for name in roboauto_options["federation"]:
        if name[:3] == coordinator_value[:3]:
            coordinator = name
            coordinator_found = True
            break
    if coordinator_found is False:
        print_err(f"coordinator {coordinator_value} not present")
        return multi_false

    return coordinator, roboauto_options["federation"][coordinator]


def roboauto_get_coordinator_from_argv(argv) -> tuple:
    """get a single coordinator from argv"""
    multi_false = False, False, False
    coordinator = False

    if len(argv) < 1:
        print_err("insert coordinator")
        return multi_false

    first_param = argv[0]
    argv = argv[1:]

    if re.match('^--', first_param) is None:
        print_err("insert coordinator")
        return multi_false

    coordinator, coordinator_url = get_coordinator_from_param(first_param)
    if coordinator is False:
        return multi_false

    return coordinator, coordinator_url, argv


def roboauto_get_multi_coordinators_from_argv(argv) -> tuple:
    """get multiple coordinators from argv, with also --all"""
    multi_false = False, False
    coordinators = []

    if len(argv) < 1:
        print_err("insert coordinators")
        return multi_false

    if argv[0] == "--all":
        return roboauto_options["federation"].keys(), argv[1:]

    while len(argv) > 0:
        first_param = argv[0]

        if re.match('^--', first_param) is None:
            break

        argv = argv[1:]

        coordinator, _ = get_coordinator_from_param(first_param)
        if coordinator is False:
            return multi_false

        coordinators.append(coordinator)

    if len(coordinators) < 1:
        print_err("insert coordinators")
        return multi_false

    return coordinators, argv


def dir_make_sure_exists(directory):
    if not os.path.exists(directory):
        if os.makedirs(directory) is not None:
            print_err("creating " + directory)
            return False
    elif not os.path.isdir(directory):
        print_err(directory + " is not a directory")
        return False

    return True


def string_is_false_none_null(string):
    # pylint: disable=R1703 simplifiable-if-statement
    if string in (
        "false", "False", "FALSE",
        "none", "None", "NONE",
        "null", "Null", "NULL"
    ):
        return True
    else:
        return False


def update_single_option(name, new_value, print_info=False):
    if roboauto_options[name] != new_value:
        old_value = roboauto_options[name]
        roboauto_options[name] = new_value
        if print_info:
            print_out(
                "option %s changed from %s to %s" %
                (name, str(old_value), str(new_value))
            )


def update_federation_option(name, new_value, print_info=False):
    if len(name) < 3:
        print_err("coordinators name should be longer than 3 letters %s not valid" % name)
        return False
    for key in roboauto_options["federation"]:
        if name != key and name[:3] == key[:3]:
            print_err("coordinator name %s not valid, similar to %s" % (name, key))
            return False

    new_value_is_none = string_is_false_none_null(new_value)

    old_value = roboauto_options["federation"].get(name, False)
    if old_value is False:
        if new_value_is_none is False:
            roboauto_options["federation"].update({name: new_value})
            if print_info:
                print_out("new coordinator %s added with url %s" % (name, new_value))
    elif old_value != new_value:
        if new_value_is_none is True:
            del roboauto_options["federation"][name]
            if print_info:
                print_out(
                    "coordinator %s deactivated old url %s" %
                    (name, str(old_value))
                )
        else:
            roboauto_options["federation"][name] = new_value
            if print_info:
                print_out(
                    "coordinator %s changed from %s to %s" %
                    (name, str(old_value), str(new_value))
                )

    return True


def update_roboauto_options(print_info=False):
    # pylint: disable=R1702 too-many-nested-blocks

    if os.path.isfile(roboauto_state["config_file"]):
        parser = configparser.RawConfigParser()
        parser.read(roboauto_state["config_file"])

        general_section = "general"
        federation_section = "federation"

        for section in list(parser):
            if section not in (
                "DEFAULT", general_section, federation_section
            ):
                print_err(f"section {section} not recognized")
                return False

        if parser.has_section(general_section):
            for option in parser.options(general_section):
                if option not in roboauto_options:
                    print_err(f"option {option} not recognied")
                    return False
                elif isinstance(roboauto_options[option], dict):
                    print_err(f"{option} is a section not an option")
                    return False

            for option in (
                "user_agent", "date_format"
            ):
                if parser.has_option(general_section, option):
                    new_value = parser.get(general_section, option).strip("'\"")
                    update_single_option(option, new_value, print_info=print_info)

            option = "create_new_after_maximum_orders"
            if parser.has_option(general_section, option):
                new_value = parser.getboolean(general_section, option)
                update_single_option(option, new_value, print_info=print_info)

            for option in (
                "seconds_pending_order", "order_maximum", "robot_maximum_orders",
                "tab_size", "routing_budget_ppm", "requests_timeout", "orders_timeout",
                "active_interval", "pending_interval", "pay_interval", "error_interval",
                "default_duration", "default_escrow"
            ):
                if parser.has_option(general_section, option):
                    try:
                        new_value = parser.getint(general_section, option)
                    except (ValueError, TypeError):
                        print_err("reading %s" % option)
                        return False

                    if option == "seconds_pending_order":
                        seconds_min_value = 300
                        if \
                            new_value != 0 and \
                            (-1 * seconds_min_value) < new_value < seconds_min_value:
                            print_err(
                                f"{option} if not 0, absolute value can "
                                f"not be less than {seconds_min_value}"
                            )
                            return False
                    elif new_value < 0:
                        print_err(f"{option} can not be negative")
                        return False

                    update_single_option(option, new_value, print_info=print_info)

            for option in (
                "default_bond_size"
            ):
                if parser.has_option(general_section, option):
                    try:
                        new_value = parser.getfloat(general_section, option)
                    except (ValueError, TypeError):
                        print_err("reading %s" % option)
                        return False

                    if new_value < 0:
                        print_err(f"{option} can not be negative")
                        return False

                    update_single_option(option, new_value, print_info=print_info)

        if parser.has_section(federation_section):
            for key in parser.options(federation_section):
                value = parser.get(federation_section, key).strip("'\"")
                if update_federation_option(key, value, print_info=print_info) is False:
                    return False

    return True


def global_setup(config_dir=None, data_dir=None):
    if config_dir is None:
        home = os.getenv("HOME")
        if home is None:
            print_err("HOME not set")
            return False
        config_home = os.getenv("XDG_CONFIG_HOME")
        if config_home is None:
            config_home = home + "/.config"
        roboauto_config = config_home + "/roboauto"
    else:
        roboauto_config = config_dir

    if data_dir is None:
        home = os.getenv("HOME")
        if home is None:
            print_err("HOME not set")
            return False
        data_home = os.getenv("XDG_DATA_HOME")
        if data_home is None:
            local_home = os.getenv("XDG_LOCAL_HOME")
            if local_home is None:
                local_home = home + "/.local"
            data_home = local_home + "/share"
        roboauto_home = data_home + "/roboauto"
    else:
        roboauto_home = data_dir

    roboauto_state["active_home"] = roboauto_home + "/active"
    roboauto_state["pending_home"] = roboauto_home + "/pending"
    roboauto_state["inactive_home"] = roboauto_home + "/inactive"
    roboauto_state["paused_home"] = roboauto_home + "/paused"
    roboauto_state["coordinators_home"] = roboauto_home + "/coordinators"
    roboauto_state["lock_home"] = roboauto_home + "/lock"
    roboauto_state["gnupg_home"] = roboauto_home + "/gnupg"
    roboauto_state["log_home"] = roboauto_home + "/logs"

    roboauto_state["waiting_queue_file"] = roboauto_home + "/waiting-queue"

    roboauto_state["config_file"] = roboauto_config + "/config.ini"
    roboauto_state["message_notification_command"] = roboauto_config + "/message-notification"
    roboauto_state["lightning_node_command"] = roboauto_config + "/lightning-node"

    for directory in (
        roboauto_home, roboauto_config,
        roboauto_state["active_home"],
        roboauto_state["pending_home"],
        roboauto_state["inactive_home"],
        roboauto_state["coordinators_home"],
        roboauto_state["paused_home"],
        roboauto_state["lock_home"],
        roboauto_state["gnupg_home"],
        roboauto_state["log_home"]
    ):
        if not dir_make_sure_exists(directory):
            return False

    all_permission = 0o777
    gnupg_home_desired_permission = 0o700
    try:
        gnupg_home_current_permission = \
            os.stat(roboauto_state["gnupg_home"]).st_mode & all_permission
        if gnupg_home_current_permission != gnupg_home_desired_permission:
            os.chmod(roboauto_state["gnupg_home"], gnupg_home_desired_permission)
    except OSError:
        print_err("changing permissions on gnupg home")
        return False

    return True


def global_shutdown():
    if roboauto_state["logger"] is not None:
        roboauto_state["logger"].close()


def state_set_command_type(command_type):
    roboauto_state["current_command_type"] = command_type


def get_int(string_number):
    try:
        number = int(string_number)
    except (ValueError, TypeError):
        print_err("%s is not a number" % string_number)
        return False

    return number


def get_uint(string_number):
    number = get_int(string_number)
    if number is False:
        return False

    if number < 0:
        print_err("%s is not positive" % number)
        return False

    return number


def is_float(float_string, additional_check : str | bool = False):
    try:
        float_float = float(float_string)
        if additional_check == "positive":
            if float_float < 0:
                return False
        elif additional_check == "negative":
            if float_float > 0:
                return False
        elif additional_check == "percentage":
            if float_float < 0 or float_float > 100:
                return False
        return True
    except (ValueError, TypeError):
        return False


def file_is_executable(file_path):
    return os.access(file_path, os.X_OK)


def json_loads(data):
    try:
        return json.loads(data)
    except json.decoder.JSONDecodeError:
        return False


def json_dumps(data):
    return json.dumps(data, indent=roboauto_options["tab_size"])


def file_write(file_name, string, error_print=True):
    try:
        with open(file_name, "w", encoding="utf8") as file:
            file.write(string + "\n")
    except EnvironmentError:
        if error_print:
            print_err("writing to %s" % file_name)
        return False

    return True


def file_json_write(file_name, data, error_print=True):
    try:
        with open(file_name, "w", encoding="utf8") as file:
            json.dump(data, file, indent=roboauto_options["tab_size"])
            file.write("\n")
    except EnvironmentError:
        if error_print:
            print_err("writing json data to %s" % file_name)
        return False

    return True


def file_read(file_name, error_print=True):
    try:
        with open(file_name, "r", encoding="utf8") as file:
            string = file.readline().rstrip()
    except EnvironmentError:
        if error_print:
            print_err("reading from %s" % file_name)
        return False

    return string


def file_json_read(file_name, error_print=True):
    try:
        with open(file_name, "r", encoding="utf8") as file:
            try:
                data = json.load(file)
            except json.decoder.JSONDecodeError:
                return False
    except EnvironmentError:
        if error_print:
            print_err("reading json data from %s" % file_name)
        return False

    return data


def file_remove(file_name):
    try:
        os.remove(file_name)
    except OSError:
        print_err(f"removing {file_name}")
        return False

    return True


def input_ask(question):
    try:
        answer = input(question)
    except KeyboardInterrupt:
        print_out("\n", end="", date=False)
        return False

    return answer


def input_ask_robot():
    return input_ask("insert robot name: ")


def password_ask(question):
    try:
        password = getpass.getpass(question)
    except KeyboardInterrupt:
        print_out("\n", end="", date=False)
        return False

    return password


def password_ask_token():
    return password_ask("insert token: ")


def list_configs():
    print_out(json_dumps(roboauto_options))

    return True


def generate_random_token_base62():
    length = 36
    characters = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789"
    random_string = ''.join(secrets.choice(characters) for _ in range(length))
    return random_string


def sha256_single(string):
    return hashlib.sha256(string.encode("utf-8")).hexdigest()


def token_get_double_sha256(token_string):
    return hashlib.sha256(
        hashlib.sha256(
            token_string.encode("utf-8")
        ).hexdigest().encode("utf-8")
    ).hexdigest()

    # return hashlib.sha256(
    #     hashlib.sha256(
    #         token_string.encode("utf-8")
    #     ).digest()
    # ).hexdigest()


# from python base91 package
def base91_encode(bindata):
    """encode a bytearray to a base91 string"""
    base91_alphabet = [
        'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M',
        'N', 'O', 'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y', 'Z',
        'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm',
        'n', 'o', 'p', 'q', 'r', 's', 't', 'u', 'v', 'w', 'x', 'y', 'z',
        '0', '1', '2', '3', '4', '5', '6', '7', '8', '9', '!', '#', '$',
        '%', '&', '(', ')', '*', '+', ',', '.', '/', ':', ';', '<', '=',
        '>', '?', '@', '[', ']', '^', '_', '`', '{', '|', '}', '~', '"'
   ]

    b = 0
    n = 0
    out = ""
    for count in range(len(bindata)):
        byte = bindata[count:count + 1]
        b |= struct.unpack("B", byte)[0] << n
        n += 8
        if n > 13:
            v = b & 8191
            if v > 88:
                b >>= 13
                n -= 13
            else:
                v = b & 16383
                b >>= 14
                n -= 14
            out += base91_alphabet[v % 91] + base91_alphabet[v // 91]
    if n:
        out += base91_alphabet[b % 91]
        if n > 7 or b > 90:
            out += base91_alphabet[b // 91]

    return out


def token_get_base91(token_string):
    return base91_encode(hashlib.sha256(token_string.encode("utf-8")).digest())


def directory_get_file_numbers(dir_number, error_print=True):
    list_number = []
    for file_number in os.listdir(dir_number):
        order_number = get_int(file_number)
        if order_number is not False:
            list_number.append(order_number)
    if len(list_number) < 1:
        return 0

    file_numbers = sorted(list_number)
    for file_number in file_numbers:
        file_path = dir_number + "/" + str(file_number)
        if not os.path.isfile(file_path):
            if error_print:
                print_err(f"{file_path} is not a file")
            return False

    return file_numbers


def string_from_multiline_format(string):
    if string is False or string is None:
        return string

    return string.replace("\\", "\n")


def string_to_multiline_format(string):
    if string is False or string is None:
        return string

    return string.replace("\n", "\\")


def lock_file_name_get(name):
    return roboauto_state["lock_home"] + "/" + name


def invoice_get_correct_amount(amount, budget_ppm):
    return int(amount * (1 - budget_ppm / 1000000))


def arg_key_value_number(var_name, arg):
    if re.match(f"^--{var_name}", arg) is None:
        return None

    key_value = arg[2:].split("=", 1)
    if len(key_value) != 2:
        print_err(f"{var_name} is not --{var_name}=number")
        return False
    key, value = key_value

    if key != var_name:
        print_err(f"key {key} not recognied")
        return False

    value_number = get_uint(value)
    if value_number is False:
        return False

    return value_number


def invoice_amount_calculate_arg(argv):
    budget_ppm = roboauto_options["routing_budget_ppm"]
    if len(argv) >= 1:
        arg_budget_ppm = arg_key_value_number("budget-ppm", argv[0])
        if arg_budget_ppm is False:
            return False
        elif arg_budget_ppm is not None:
            budget_ppm = arg_budget_ppm
            argv = argv[1:]

    if len(argv) < 1:
        print_err("insert full-amount")
        return False

    full_amount = get_int(argv[0])
    if full_amount is False:
        return False
    argv = argv[1:]

    correct_amount = invoice_get_correct_amount(full_amount, budget_ppm)

    print_out(correct_amount)

    return True


def random_interval(max_value):
    return random.randint(1, max_value) - 1


def shuffle_array(ordered: list) -> list:
    shuffled = ordered.copy()
    random.shuffle(shuffled)
    return shuffled


def shuffle_dic(ordered: dict) -> dict:
    keys = list(ordered.keys())
    random.shuffle(keys)
    shuffled = {}
    for key in keys:
        shuffled[key] = ordered[key]
    return shuffled


def bad_request_is_cancelled(bad_request):
    # may be changed in the future
    # https://github.com/RoboSats/robosats/issues/1245

    if not isinstance(bad_request, str):
        return False

    return bad_request in (
        "This order has been cancelled by the maker",
        "This order has been cancelled collaborativelly"
    )


def bad_request_is_wrong_robot(bad_request):
    if not isinstance(bad_request, str):
        return False

    # pylint: disable=C0301 line-too-long
    return bad_request in (
        "Robot token SHA256 was provided in the header. However it is not a valid 39 or 40 characters Base91 string."
        "On the first request to a RoboSats coordinator, you must provide as well a valid public and encrypted private PGP keys"
    )



#!/usr/bin/env python3

"""order_data.py"""

# pylint: disable=C0116 missing-function-docstring


def get_type_string(target, reverse=False):
    types = {
        0: "buy",
        1: "sell",
        2: "all"
    }

    if not reverse:
        if target in types:
            return types[target]
        else:
            return "other"
    else:
        for type_id, type_string in types.items():
            if target == type_string:
                return type_id
        return -1


def get_all_currencies():
    return {
           1: "USD",
           2: "EUR",
           3: "JPY",
           4: "GBP",
           5: "AUD",
           6: "CAD",
           7: "CHF",
           8: "CNY",
           9: "HKD",
          10: "NZD",
          11: "SEK",
          12: "KRW",
          13: "SGD",
          14: "NOK",
          15: "MXN",
          16: "BYN",
          17: "RUB",
          18: "ZAR",
          19: "TRY",
          20: "BRL",
          21: "CLP",
          22: "CZK",
          23: "DKK",
          24: "HRK",
          25: "HUF",
          26: "INR",
          27: "ISK",
          28: "PLN",
          29: "RON",
          30: "ARS",
          31: "VES",
          32: "COP",
          33: "PEN",
          34: "UYU",
          35: "PYG",
          36: "BOB",
          37: "IDR",
          38: "ANG",
          39: "CRC",
          40: "CUP",
          41: "DOP",
          42: "GHS",
          43: "GTQ",
          44: "ILS",
          45: "JMD",
          46: "KES",
          47: "KZT",
          48: "MYR",
          49: "NAD",
          50: "NGN",
          51: "AZN",
          52: "PAB",
          53: "PHP",
          54: "PKR",
          55: "QAR",
          56: "SAR",
          57: "THB",
          58: "TTD",
          59: "VND",
          60: "XOF",
          61: "TWD",
          62: "TZS",
          63: "XAF",
          64: "UAH",
          65: "EGP",
          66: "LKR",
          67: "MAD",
          68: "AED",
          69: "TND",
          70: "ETB",
          71: "GEL",
          72: "UGX",
          73: "RSD",
          74: "IRT",
          75: "BDT",
          76: "ALL",
         300: "XAU",
        1000: "BTC"
    }


def get_currency_string(target, reverse=False):
    currencies = get_all_currencies()
    if not reverse:
        if target in currencies:
            return currencies[target]
        else:
            return "???"
    else:
        for currency_id, currency_string in currencies.items():
            if target.upper() == currency_string:
                return currency_id
        return -1


def get_fiat_payment_methods():
    return [
        "Revolut",
        "CashApp",
        "Zelle",
        "Strike",
        "WeChat Pay",
        "Instant SEPA",
        "Interac e-Transfer",
        "Wise",
        "Venmo",
        "Faster Payments",
        "Paypal Friends & Family",
        "LINE Pay",
        "Rakuten Pay",
        "PromptPay",
        "Bizum",
        "N26",
        "Tinkoff Bank",
        "TWINT",
        "Advcash",
        "BLIK",
        "MBWay",
        "W1TTY",
        "Verse",
        "Paysera",
        "Amazon GiftCard",
        "Ozon GiftCard",
        "AliPay",
        "GPay",
        "Bancolombia",
        "Pago Movil BDV",
        "SPEI",
        "PIX",
        "Pouch.ph",
        "PayID",
        "Paysafe",
        "Sber Bank",
        "PhonePe",
        "OVO",
        "HalCash",
        "Vivid",
        "Google Play Gift Code",
        "Apple Pay",
        "Steam",
        "Nequi",
        "ShakePay",
        "Sinpe",
        "DaviPlata",
        "CoDi",
        "TaiwanPay",
        "MaiCoin",
        "GoPay",
        "MercadoPago",
        "Monero",
        "USDT",
        "L-USDt",
        "Airtel Money",
        "MTN Money",
        "M-Pesa",
        "UPI",
        "MoMo",
        "Tigo Pesa",
        "Cash F2F",
        "Amazon USA GiftCard",
        "Amazon DE GiftCard",
        "Amazon AU GiftCard",
        "Amazon SA GiftCard",
        "Amazon ES GiftCard",
        "Amazon CA GiftCard",
        "Amazon CN GiftCard",
        "Amazon AE GiftCard",
        "Amazon FR GiftCard",
        "Amazon NL GiftCard",
        "Amazon IN GiftCard",
        "Amazon IT GiftCard",
        "Amazon JP GiftCard",
        "Amazon MX GiftCard",
        "Amazon PL GiftCard",
        "Amazon UK GiftCard",
        "Amazon SE GiftCard",
        "Amazon SG GiftCard",
        "Amazon TR GiftCard",
        "Tinkoff QR",
        "SBP",
        "Qiwi"
    ]


def get_swap_payment_methods():
    return [
        "On-Chain BTC",
        "On-Chain w/ Stowaway",
        "RBTC",
        "LBTC",
        "WBTC"
    ]


def get_order_string(target, reverse=False):
    status_dic = {
         0: "Waiting for maker bond",
         1: "Public",
         2: "Paused",
         3: "Waiting for taker bond",
         4: "Cancelled",
         5: "Expired",
         6: "Waiting for trade collateral and buyer invoice",
         7: "Waiting only for seller trade collateral",
         8: "Waiting only for buyer invoice",
         9: "Sending fiat - In chatroom",
        10: "Fiat sent - In chatroom",
        11: "In dispute",
        12: "Collaboratively cancelled",
        13: "Sending satoshis to buyer",
        14: "Sucessful trade",
        15: "Failed lightning network routing",
        16: "Wait for dispute resolution",
        17: "Maker lost dispute",
        18: "Taker lost dispute"
    }

    if not reverse:
        if target in status_dic:
            return status_dic[target]
        else:
            return "other"
    else:
        for status_id, status_string in status_dic.items():
            if target == status_string:
                return status_id
        return -1


def order_is_waiting_maker_bond(data):
    return data == 0


def order_is_public(data):
    return data == 1


def order_is_paused(data):
    return data == 2


def order_is_waiting_taker_bond(data):
    return data == 3


def order_is_cancelled(data):
    return data == 4


def order_is_expired(data):
    return data == 5


def order_is_waiting_seller_buyer(data):
    return data == 6


def order_is_waiting_seller(data):
    return data == 7


def order_is_waiting_buyer(data):
    return data == 8


def order_is_waiting_fiat_sent(data):
    return data == 9


def order_is_fiat_sent(data):
    return data == 10


def order_is_in_dispute(data):
    return data == 11


def order_is_collaborative_cancelled(data):
    return data == 12


def order_is_sending_to_buyer(data):
    return data == 13


def order_is_sucessful(data):
    return data == 14


def order_is_failed_routing(data):
    return data == 15


def order_is_waiting_dispute_resolution(data):
    return data == 16


def order_is_maker_lost_dispute(data):
    return data == 17


def order_is_taker_lost_dispute(data):
    return data == 18


def order_is_pending(data):
    return data in (6, 7, 8, 9, 10, 11, 13, 15, 16)


def order_is_finished(data):
    return data in (4, 12, 14, 17, 18)


def order_is_finished_for_seller(data):
    return data in (13, 15)


# robosats/api/models/order.py
def get_order_expiry_reason_string(target, reverse=False):
    expiry_reasons_dic = {
        0: "Expired not taken",
        1: "Maker bond not locked",
        2: "Escrow not locked",
        3: "Invoice not submitted",
        4: "Neither escrow locked or invoice submitted"
    }

    if not reverse:
        if target in expiry_reasons_dic:
            return expiry_reasons_dic[target]
        else:
            return "other"
    else:
        for expiry_reason, expiry_message in expiry_reasons_dic.items():
            if target == expiry_message:
                return expiry_reason
        return -1


def order_expired_is_not_taken(data):
    return data == 0


def order_expired_is_maker_bond_not_locked(data):
    return data == 1



#!/usr/bin/env python3

"""order_action.py"""

# pylint: disable=C0116 missing-function-docstring

import os

from roboauto.global_state import roboauto_options, roboauto_state
from roboauto.utils import \
    print_out, print_err, get_uint, json_loads, json_dumps, \
    input_ask, file_write, file_read, arg_key_value_number, \
    invoice_get_correct_amount, is_float, file_is_executable
from roboauto.robot import \
    robot_get_current_fingerprint, robot_var_from_dic, \
    robot_input_from_argv, robot_change_dir
from roboauto.order_data import \
    order_is_waiting_seller_buyer, order_is_waiting_buyer, \
    order_is_waiting_seller, order_is_waiting_fiat_sent, \
    order_is_fiat_sent, order_is_public, order_is_paused, \
    order_is_sucessful, order_is_expired, get_order_string, \
    order_is_in_dispute, order_is_failed_routing, \
    order_is_sending_to_buyer
from roboauto.order import \
    order_requests_order_dic, peer_nick_from_response, bond_order, \
    amount_correct_from_response, premium_string_get, \
    order_string_status_print
from roboauto.requests_api import \
    requests_api_order_invoice, requests_api_order_pause, \
    requests_api_order_confirm, requests_api_order_undo_confirm, \
    requests_api_order_dispute, response_is_error, \
    requests_api_order_cancel, requests_api_order_rate, \
    requests_api_order_address, requests_api_order_submit_statement
from roboauto.gpg_key import gpg_sign_message
from roboauto.subprocess_commands import \
    subprocess_generate_invoice, \
    subprocess_pay_invoice_and_check
from roboauto.chat import robot_requests_chat


def order_take_argv(argv):
    if len(argv) < 1:
        print_err("insert options")
        return False

    fully = False
    use_node = True
    while len(argv) >= 1:
        if argv[0] == "--fully":
            argv = argv[1:]
            fully = True
        elif argv[0] == "--no-node":
            argv = argv[1:]
            use_node = False
        else:
            break

    if fully and not use_node:
        print_err("--fully and --no-node should not both be present")
        return False

    if not file_is_executable(roboauto_state["lightning_node_command"]) and fully is True:
        print_err("lightning node is not set --fully can not be used")
        return False

    robot_dic, argv = robot_input_from_argv(argv)
    if robot_dic is False:
        return False

    if len(argv) >= 1:
        order_id = argv[0]
        argv = argv[1:]
    else:
        order_id = input_ask("insert order id: ")
        if order_id is False:
            return False
    if get_uint(order_id) is False:
        return False

    if len(argv) >= 1:
        take_amount = argv[0]
        argv = argv[1:]
        if get_uint(take_amount) is False:
            return False
    else:
        take_amount = None

    order_dic = bond_order(
        robot_dic, order_id, taker=True,
        take_amount=take_amount,
        use_node=use_node
    )
    if order_dic is False:
        return False

    if use_node is False:
        return True

    robot_name = robot_dic["name"]

    print_out(f"{robot_name} {order_id} taken successfully")

    if fully is True:
        if isinstance(order_dic, dict):
            if order_dic["order_response_json"].get("is_seller", False) is True:
                if order_seller_bond_escrow(robot_dic, True) is False:
                    return False
            else:
                if order_buyer_update_invoice(robot_dic, None) is False:
                    return False

    if not robot_change_dir(robot_name, "pending"):
        return False

    return True


def order_buyer_update_data(
    robot_dic, data_string, data_var, extra_value
):
    robot_name, _, robot_dir, token, _, token_base91, robot_url = robot_var_from_dic(robot_dic)

    order_dic = order_requests_order_dic(robot_dic, order_id=False)
    if order_dic is False or isinstance(order_dic, str):
        return False

    order_user = order_dic["order_user"]
    order_info = order_dic["order_info"]
    order_response_json = order_dic["order_response_json"]

    order_id = order_info["order_id"]
    status_id = order_info["status"]

    is_buyer = order_response_json.get("is_buyer", False)
    if is_buyer is False:
        print_err(f"{robot_name} {order_id} is not buyer")
        return False

    if \
        not order_is_waiting_seller_buyer(status_id) and \
        not order_is_waiting_buyer(status_id):
        if data_string == "invoice":
            if not order_is_failed_routing(status_id):
                print_err(
                    f"{robot_name} {order_id} is not "
                    "waiting for buyer or failed routing"
                )
                return False
        else:
            print_err(f"{robot_name} {order_id} is not waiting for buyer")
            return False

    order_description = order_info["order_description"]

    peer_nick = peer_nick_from_response(order_response_json)

    order_string_status_print(robot_name, order_id, order_description, peer_nick)

    if data_string == "invoice":
        invoice = data_var
        budget_ppm = extra_value

        if invoice is None or invoice is False:
            invoice_amount = order_response_json.get("trade_satoshis", False)
            if invoice_amount is False:
                print_err("trade satoshis is not present")
                return False

            correct_invoice_amount = invoice_get_correct_amount(invoice_amount, budget_ppm)

            amount_correct = amount_correct_from_response(order_response_json)
            if amount_correct is False:
                amount_correct = order_info["amount_string"]

            invoice = subprocess_generate_invoice(
                str(correct_invoice_amount),
                robot_name + "-" + str(peer_nick) + "-" + order_id + "-" +
                order_user["type"] + "-" + order_user["currency"] + "-" +
                amount_correct + "-" + premium_string_get(order_user["premium"])
            )
            if invoice is False:
                return False

        message = invoice
        requests_api_order_function = requests_api_order_invoice
    elif data_string == "address":
        address = data_var
        message = address
        requests_api_order_function = requests_api_order_address
    else:
        print_err(f"wrong order buyer update data {data_string}")
        return False

    fingerprint = robot_get_current_fingerprint(robot_dir)
    if fingerprint is False:
        return False

    signed_message = gpg_sign_message(message, fingerprint, passphrase=token)
    if signed_message is False:
        print_err(f"signing {data_string}")
        return False

    order_data_response_all = requests_api_order_function(
        token_base91, order_id, robot_url, robot_name, signed_message, extra_value
    )
    if response_is_error(order_data_response_all):
        return False
    order_data_response = order_data_response_all.text
    order_data_response_json = json_loads(order_data_response)
    if order_data_response_json is False:
        print_err(order_data_response, end="", error=False, date=False)
        print_err(f"{robot_name} {order_id} sending {data_string}")
        return False

    bad_request = order_data_response_json.get("bad_request", False)
    if bad_request is not False:
        print_err(bad_request, date=False, error=False)
        return False

    new_status_id = order_data_response_json.get("status", False)
    if new_status_id is False:
        print_err(json_dumps(order_data_response_json), date=False, error=False)
        print_err(f"{data_string} not sent, no new status")
        return False

    if \
        not order_is_waiting_seller(new_status_id) and \
        not order_is_waiting_fiat_sent(new_status_id) and \
        not order_is_sending_to_buyer(new_status_id):
        print_err(json_dumps(order_data_response_json), date=False, error=False)
        print_err(
            f"{data_string} not send, current status: " +
            str(get_order_string(new_status_id))
        )
        return False

    print_out(f"{robot_name} {order_id} {data_string} sent successfully")

    return True


def order_buyer_update_invoice(robot_dic, extra_arg):
    budget_ppm, invoice = extra_arg
    if budget_ppm is None or budget_ppm is False:
        budget_ppm = roboauto_options["routing_budget_ppm"]

    return order_buyer_update_data(
        robot_dic, "invoice", invoice, budget_ppm
    )


def order_buyer_update_address(robot_dic, extra_arg):
    address, sat_per_vb = extra_arg

    return order_buyer_update_data(
        robot_dic, "address", address, sat_per_vb
    )


def order_seller_bond_escrow(robot_dic, extra_arg):
    robot_name = robot_dic["name"]

    order_dic = order_requests_order_dic(robot_dic, order_id=False)
    if order_dic is False or isinstance(order_dic, str):
        return False

    # pylint: disable=R0801 duplicate-code
    order_user = order_dic["order_user"]
    order_info = order_dic["order_info"]
    order_response_json = order_dic["order_response_json"]

    order_id = order_info["order_id"]
    status_id = order_info["status"]

    is_seller = order_response_json.get("is_seller", False)
    if is_seller is False:
        print_err(f"{robot_name} {order_id} is not seller")
        return False

    if \
        not order_is_waiting_seller_buyer(status_id) and \
        not order_is_waiting_seller(status_id):
        print_err(f"{robot_name} {order_id} is not waiting for seller")
        return False

    order_description = order_info["order_description"]

    peer_nick = peer_nick_from_response(order_response_json)

    order_string_status_print(robot_name, order_id, order_description, peer_nick)

    amount_correct = amount_correct_from_response(order_response_json)
    if amount_correct is False:
        amount_correct = order_info["amount_string"]

    escrow_satoshis = order_response_json.get("escrow_satoshis", False)
    if escrow_satoshis is False:
        print_err(
            f"{robot_name} {order_id} escrow_satoshis not present, invoice can not be checked"
        )
        return False

    escrow_invoice = order_response_json.get("escrow_invoice", False)
    if escrow_invoice is False:
        print_err(
            f"{robot_name} {order_id} escrow_invoice not present, invoice can not be checked"
        )
        return False

    use_node = extra_arg
    if not use_node:
        print_out(escrow_invoice)
        return True

    if not file_is_executable(roboauto_state["lightning_node_command"]):
        print_err("lightning node not set, to use without node pass --no-node")
        return False

    pay_label = \
        robot_name + "-" + str(peer_nick) + "-" + order_id + "-" + \
        order_user["type"] + "-" + order_user["currency"] + "-" + \
        amount_correct + "-" + premium_string_get(order_user["premium"])
    if subprocess_pay_invoice_and_check(
        robot_dic, order_id,
        escrow_invoice, str(escrow_satoshis), pay_label,
        lambda order_status : \
            not order_is_waiting_seller_buyer(order_status) and \
            not order_is_waiting_seller(order_status),
        "checking if escrow is paid...",
        "escrow paid successfully",
        "escrow not paid in time",
        order_requests_order_dic, order_is_expired,
        maximum_retries=100
    ) is False:
        return False

    return True


def order_post_action_simple(
    robot_dic, order_post_function, is_wrong_status,
    string_error, string_or_bad_request,
    extra_arg=None,
    should_be_buyer=False, should_be_seller=False
):
    robot_name, _, _, _, _, token_base91, robot_url = robot_var_from_dic(robot_dic)

    order_dic = order_requests_order_dic(robot_dic, order_id=False)
    if order_dic is False or isinstance(order_dic, str):
        return False

    order_info = order_dic["order_info"]
    order_response_json = order_dic["order_response_json"]

    status_id = order_info["status"]
    order_id = order_info["order_id"]

    if should_be_buyer and order_response_json.get("is_seller", False):
        print_err(f"{robot_name} {order_id} is not buyer")
        return False

    if should_be_seller and order_response_json.get("is_buyer", False):
        print_err(f"{robot_name} {order_id} is not seller")
        return False

    if is_wrong_status(status_id):
        print_err(f"{robot_name} {order_id} " + string_error)
        return False

    order_description = order_info["order_description"]

    peer_nick = peer_nick_from_response(order_response_json)

    order_string_status_print(robot_name, order_id, order_description, peer_nick)

    if extra_arg is False or extra_arg is None:
        order_post_response_all = order_post_function(
            token_base91, order_id, robot_url, robot_name
        )
    else:
        order_post_response_all = order_post_function(
            token_base91, order_id, robot_url, extra_arg
        )
    if response_is_error(order_post_response_all):
        return False
    order_post_response = order_post_response_all.text
    order_post_response_json = json_loads(order_post_response)
    if order_post_response_json is False:
        print_err(order_post_response, error=False, date=False)
        print_err(f"{robot_name} {order_id} response is not json")
        return False

    # strange reseponse for order post cancel
    # bad_request is set when success
    # see https://github.com/RoboSats/robosats/issues/1245
    bad_request = order_post_response_json.get("bad_request", False)
    if string_or_bad_request is False:
        if bad_request is False:
            print_err(order_post_response, error=False, date=False)
            print_err(f"{robot_name} {order_id} problems in the response")
            return False

        print_out(bad_request, date=False)
    else:
        if bad_request is not False:
            print_err(bad_request, error=False, date=False)
            print_err(f"{robot_name} {order_id} problems in the response")
            return False

        print_out(f"{robot_name} {order_id} " + string_or_bad_request)

    return True


def order_pause_toggle(robot_dic):
    return order_post_action_simple(
        robot_dic,
        requests_api_order_pause,
        lambda status_id : \
            not order_is_public(status_id) and \
            not order_is_paused(status_id),
        "is not public or paused",
        "toggled pause"
    )


def order_collaborative_cancel(robot_dic):
    return order_post_action_simple(
        robot_dic,
        requests_api_order_cancel,
        lambda status_id : \
            not order_is_waiting_buyer(status_id) and \
            not order_is_waiting_fiat_sent(status_id),
        "can not send collaborative cancel",
        False
    )


def order_send_confirm(robot_dic):
    return order_post_action_simple(
        robot_dic,
        requests_api_order_confirm,
        lambda status_id : \
            not order_is_waiting_fiat_sent(status_id) and \
            not order_is_fiat_sent(status_id),
        "can not confirm payment",
        "confirmation sent"
    )


def order_undo_confirm(robot_dic):
    return order_post_action_simple(
        robot_dic,
        requests_api_order_undo_confirm,
        lambda status_id : \
            not order_is_fiat_sent(status_id),
        "can not undo confirm payment",
        "confirmation undone",
        should_be_buyer=True
    )


def order_start_dispute(robot_dic):
    return order_post_action_simple(
        robot_dic,
        requests_api_order_dispute,
        lambda status_id : \
            not order_is_waiting_fiat_sent(status_id) and \
            not order_is_fiat_sent(status_id),
        "can not start dispute",
        "dispute started"
    )


def order_submit_statement(robot_dic, extra_arg):
    statement = extra_arg

    return order_post_action_simple(
        robot_dic,
        requests_api_order_submit_statement,
        lambda status_id : \
            not order_is_in_dispute(status_id),
        "can not send dispute statement",
        "dispute statement submitted",
        extra_arg=statement
    )


def order_rate_coordinator(robot_dic, extra_arg):
    rating = extra_arg

    if order_post_action_simple(
        robot_dic,
        requests_api_order_rate,
        lambda status_id : \
            not order_is_sucessful(status_id),
        "is not sucessful",
        f"robosats rated {rating} stars",
        extra_arg=rating
    ) is False:
        return False

    if not file_write(robot_dic["dir"] + "/rating-coordinator", rating):
        return False

    return True


def robot_order_post_action_argv(argv, order_post_function, extra_type=None):
    extra_arg = ()

    if extra_type == "escrow_pay":
        use_node = True
        if len(argv) >= 1 and argv[0] == "--no-node":
            use_node = False
            argv = argv[1:]

        extra_arg = use_node

    # pylint: disable=R0801 duplicate-code
    robot_dic, argv = robot_input_from_argv(argv)
    if robot_dic is False:
        return False

    if extra_type == "update_invoice":
        budget_ppm = None
        if len(argv) >= 1:
            arg_budget_ppm = arg_key_value_number("budget-ppm", argv[0])
            if arg_budget_ppm is False:
                return False
            elif arg_budget_ppm is not None:
                budget_ppm = arg_budget_ppm
                argv = argv[1:]

        invoice = None
        if len(argv) >= 1:
            invoice = argv[0]
            argv = argv[1:]
        if \
            not file_is_executable(roboauto_state["lightning_node_command"]) and \
            invoice is None:
            invoice = input_ask("insert invoice: ")
            if invoice is False:
                return False
        if invoice == "":
            print_err("invoice not set")

        extra_arg = (budget_ppm, invoice)
    elif extra_type == "update_address":
        if len(argv) < 1:
            print_err("insert address")
            return False
        address = argv[0]
        argv = argv[1:]

        if len(argv) < 1:
            print_err("insert sat per vb")
            return False
        sat_per_vb = argv[0]
        argv = argv[1:]

        if not is_float(sat_per_vb, additional_check="positive"):
            print_err("sat per vb should be a positive float")
            return False

        extra_arg = (address, sat_per_vb)
    elif extra_type == "statement":
        # see frontend/src/components/TradeBox/index.tsx
        # maybe support also sending chat messages, not working in the main web client
        if len(argv) < 1:
            print_err("insert dispute statement")
            return False
        statement_arg = argv[0]
        argv = argv[1:]

        if statement_arg != "--file":
            statement = statement_arg
        else:
            if len(argv) < 1:
                print_err("insert dispute statement file")
                return False
            statement_file = argv[0]
            argv = argv[1:]

            if not os.path.isfile(statement_file):
                print_err(f"{statement_file} is not a file")
                return False
            statement = file_read(statement_file)
            if statement is False:
                return False

        statement_min_len = 100
        statement_max_len = 5000
        if len(statement) <= statement_min_len:
            print_err(f"dispute statement should be at least {statement_min_len} characters long")
            return False
        elif len(statement) >= statement_max_len:
            print_err(f"dispute statement should be at most {statement_max_len} characters long")
            return False

        extra_arg = statement
    elif extra_type == "rating":
        robot_name = robot_dic["name"]
        rating_coordinator_file = robot_dic["dir"] + "/rating-coordinator"
        if os.path.isfile(rating_coordinator_file):
            rating = file_read(rating_coordinator_file)
            if rating is False:
                return False
            print_err(f"{robot_name} already rated {rating}")
            return False

        if len(argv) < 1:
            print_err("insert rating")
            return False
        rating = argv[0]
        argv = argv[1:]

        rating_uint = get_uint(rating)
        if rating_uint is False:
            return False
        if rating_uint < 1 or rating_uint > 5:
            print_err("rating should be between 1 and 5")
            return False

        extra_arg = rating
    elif extra_type == "save_chat":
        save_chat = True
        if len(argv) >= 1 and argv[0] == "--no-save-chat":
            argv = argv[1:]
            save_chat = False

        if save_chat is True:
            chat_response, chat_response_json = robot_requests_chat(robot_dic)
            if chat_response is False or chat_response_json is False:
                return False
            robot_name = robot_dic["name"]
            print_out(f"{robot_name} chat saved on disk")

    if extra_arg == ():
        return order_post_function(robot_dic)
    else:
        return order_post_function(robot_dic, extra_arg)


def order_buyer_update_invoice_argv(argv):
    return robot_order_post_action_argv(
        argv, order_buyer_update_invoice, extra_type="update_invoice"
    )


def order_buyer_update_address_argv(argv):
    return robot_order_post_action_argv(
        argv, order_buyer_update_address, extra_type="update_address"
    )


def order_submit_statement_argv(argv):
    return robot_order_post_action_argv(
        argv, order_submit_statement, extra_type="statement"
    )


def order_rate_coordinator_argv(argv):
    return robot_order_post_action_argv(
        argv, order_rate_coordinator, extra_type="rating"
    )


def order_collaborative_cancel_argv(argv):
    return robot_order_post_action_argv(
        argv, order_collaborative_cancel, extra_type="save_chat"
    )


def order_send_confirm_argv(argv):
    return robot_order_post_action_argv(
        argv, order_send_confirm, extra_type="save_chat"
    )


def order_start_dispute_argv(argv):
    return robot_order_post_action_argv(
        argv, order_start_dispute, extra_type="save_chat"
    )


def order_seller_bond_escrow_argv(argv):
    return robot_order_post_action_argv(
        argv, order_seller_bond_escrow, extra_type="escrow_pay"
    )


def order_pause_toggle_argv(argv):
    return robot_order_post_action_argv(argv, order_pause_toggle)


def order_undo_confirm_argv(argv):
    return robot_order_post_action_argv(argv, order_undo_confirm)



#!/usr/bin/env python3

"""info.py"""

# pylint: disable=C0116 missing-function-docstring

import sys
import os
import re

from roboauto.logger import print_out, print_err
from roboauto.robot import \
    robot_input_from_argv, robot_requests_robot, \
    robot_var_from_dic, robot_requests_get_order_id
from roboauto.order_local import \
    order_robot_get_last_order_id, order_dic_from_robot_dir, \
    order_dic_print
from roboauto.order import order_requests_order_dic
from roboauto.chat import \
    robot_requests_chat, chat_print_encrypted_messages, chat_print_single_message
from roboauto.requests_api import \
    response_is_error, requests_api_info, \
    requests_api_historical, requests_api_limits, \
    requests_api_price, requests_api_ticks
from roboauto.utils import \
    file_json_read, json_loads, json_dumps, \
    roboauto_get_coordinator_url, roboauto_get_coordinator_from_argv, \
    token_get_base91


def requests_simple_handle(
    requests_function, coordinator_url, coordinator, description_string
):
    response_all = requests_function(coordinator_url, coordinator)
    if response_is_error(response_all):
        return False
    response = response_all.text
    response_json = json_loads(response)
    if response_json is False:
        print_err(response, end="", error=False, date=False)
        print_err(f"{description_string} response is not json")
        return False

    return response_json


def list_historical(argv):
    coordinator, coordinator_url, argv = roboauto_get_coordinator_from_argv(argv)
    if coordinator_url is False:
        return False

    historical_response_json = requests_simple_handle(
        requests_api_historical, coordinator_url, coordinator, "historical"
    )
    if historical_response_json is False:
        return False

    print_out(json_dumps(historical_response_json))

    return True


def list_limits(argv):
    coordinator, coordinator_url, argv = roboauto_get_coordinator_from_argv(argv)
    if coordinator_url is False:
        return False

    limits_response_json = requests_simple_handle(
        requests_api_limits, coordinator_url, coordinator, "limits"
    )
    if limits_response_json is False:
        return False

    print_out(json_dumps(limits_response_json))

    return True


def list_price(argv):
    coordinator, coordinator_url, argv = roboauto_get_coordinator_from_argv(argv)
    if coordinator_url is False:
        return False

    price_response_json = requests_simple_handle(
        requests_api_price, coordinator_url, coordinator, "price"
    )
    if price_response_json is False:
        return False

    print_out(json_dumps(price_response_json))

    return True


def list_ticks(argv: list):
    coordinator, coordinator_url, argv = roboauto_get_coordinator_from_argv(argv)
    if coordinator_url is False:
        return False

    if len(argv) < 1:
        print_err("insert start date")
        return False
    start_date = argv[0]
    argv = argv[1:]

    if len(argv) < 1:
        print_err("insert end date")
        return False
    end_date = argv[0]
    argv = argv[1:]

    ticks_response_all = requests_api_ticks(
        coordinator_url, coordinator, start_date, end_date
    )
    if response_is_error(ticks_response_all):
        return False
    ticks_response = ticks_response_all.text
    ticks_response_json = json_loads(ticks_response)
    if ticks_response_json is False:
        print_err(ticks_response, end="", error=False, date=False)
        print_err("ticks response is not json")
        return False

    print_out(json_dumps(ticks_response_json))

    return True


def robosats_info(argv):
    coordinator, coordinator_url, argv = roboauto_get_coordinator_from_argv(argv)
    if coordinator_url is False:
        return False

    info_response_json = requests_simple_handle(
        requests_api_info, coordinator_url, coordinator, "info"
    )
    if info_response_json is False:
        return False

    print_out(json_dumps(info_response_json))

    return True


def robot_info_argv(argv):
    """print info about a robot"""

    robot_dic = None

    token_base91 = False

    while len(argv) > 0:
        if argv[0] == "--stdin":
            token_base91 = token_get_base91(sys.stdin.readline().rstrip())
        elif argv[0] == "--stdin-base91":
            token_base91 = sys.stdin.readline().rstrip()
        else:
            break
        argv = argv[1:]

    if token_base91 is False:
        robot_dic, argv = robot_input_from_argv(argv)
        if robot_dic is False:
            return False

        token_base91 = token_get_base91(robot_dic["token"])
        robot_url = roboauto_get_coordinator_url(robot_dic["coordinator"])
    else:
        if len(argv) < 1:
            print_err("insert coordinator name or link")
            return False

        if re.match('^--', argv[0]) is None:
            robot_url = argv[0]
            argv = argv[1:]
        else:
            _, robot_url, argv = roboauto_get_coordinator_from_argv(argv)
            if robot_url is False:
                return False

    # pylint: disable=R0801 duplicate-code
    robot_response, robot_response_json = robot_requests_robot(
        token_base91, robot_url, robot_dic
    )
    if robot_response is False:
        return False

    print_out(json_dumps(robot_response_json))

    return True


def order_info_argv(argv):
    local_mode = False
    search_mode = False
    full_mode = True
    while len(argv) > 0:
        first_argv = argv[0]
        if first_argv == "--local":
            local_mode = True
            argv = argv[1:]
        elif first_argv == "--search":
            search_mode = True
            argv = argv[1:]
        elif first_argv == "--simple":
            full_mode = False
            argv = argv[1:]
        elif re.match('^-', first_argv) is not None:
            print_err(f"{first_argv} not recognied")
            return False
        else:
            break

    # pylint: disable=R0801 duplicate-code
    robot_dic, argv = robot_input_from_argv(argv)
    if robot_dic is False:
        return False

    robot_name, _, robot_dir, _, coordinator, _, _ = robot_var_from_dic(robot_dic)

    if len(argv) >= 1:
        if search_mode is True:
            print_err("order_id should not be specified with --search")
            return False
        order_id = argv[0]
        argv = argv[1:]
    else:
        order_id = False

    if local_mode is False:
        if order_id is False:
            if search_mode is False:
                order_id = order_robot_get_last_order_id(robot_dic, error_print=False)
                if order_id is False:
                    print_err("robot does not have orders saved, searching it", error=False)

            if order_id is False:
                order_id = robot_requests_get_order_id(robot_dic, error_print=False)
                if order_id is False:
                    print_err(f"{robot_name} does not have active or last orders")
                    return False

        order_dic = order_requests_order_dic(robot_dic, order_id)
        if order_dic is False or isinstance(order_dic, str):
            return False
    else:
        order_dic = order_dic_from_robot_dir(
            robot_dir, order_id=order_id, error_print=False
        )
        if order_dic is None:
            print_err(json_dumps({"error": "no order dir"}))
            return False
        if order_dic is False:
            return False

    order_dic_print(
        order_dic, robot_name, coordinator, one_line=False, full_mode=full_mode
    )

    return True


def robot_chat(argv):
    from_local = False
    if len(argv) > 0 and argv[0] == "--local":
        from_local = True
        argv = argv[1:]

    robot_dic, argv = robot_input_from_argv(argv)
    if robot_dic is False:
        return False

    robot_dir = robot_dic["dir"]

    if from_local is False:
        chat_response, chat_response_json = robot_requests_chat(robot_dic)
        if chat_response is False:
            return False

        if not chat_print_encrypted_messages(robot_dic, chat_response_json):
            return False
    else:
        decrypted_messages_file = robot_dir + "/messages-decrypted"
        chat_response_file = robot_dir + "/chat-response"

        if os.path.isfile(decrypted_messages_file):
            decrypted_messages = file_json_read(decrypted_messages_file)
            if decrypted_messages is False:
                return False

            first_message = True
            for message_dic in decrypted_messages:
                if first_message:
                    first_message = False
                else:
                    print_out("\n", end="")

                if chat_print_single_message(message_dic) is False:
                    return False
        elif os.path.isfile(chat_response_file):
            chat_response_json = file_json_read(chat_response_file)
            if chat_response_json is False:
                return False

            if not chat_print_encrypted_messages(robot_dic, chat_response_json):
                return False
        else:
            print_err("there are no local messages")
            return False

    return True



#!/usr/bin/env python3

"""chat.py"""

# pylint: disable=C0116 missing-function-docstring

import re

from roboauto.logger import print_out, print_err
from roboauto.date_utils import date_convert_time_zone_and_format_string
from roboauto.utils import \
    json_loads, string_from_multiline_format, string_to_multiline_format, \
    input_ask, file_json_write
from roboauto.order_local import order_robot_get_last_order_id
from roboauto.requests_api import \
    response_is_error, requests_api_chat_post, requests_api_chat
from roboauto.robot import \
    robot_save_peer_gpg_public_key, robot_get_current_fingerprint, \
    robot_get_peer_fingerprint, robot_input_from_argv, robot_var_from_dic
from roboauto.gpg_key import \
    gpg_import_key, gpg_encrypt_sign_message, gpg_decrypt_check_message


def robot_requests_chat(robot_dic):
    multi_false = False, False

    robot_name, _, robot_dir, _, _, token_base91, robot_url = robot_var_from_dic(robot_dic)

    order_id = order_robot_get_last_order_id(robot_dic)
    if order_id is False:
        return multi_false

    chat_response_all = requests_api_chat(token_base91, order_id, robot_url, robot_name)
    if response_is_error(chat_response_all):
        return multi_false
    chat_response = chat_response_all.text
    chat_response_json = json_loads(chat_response)
    if chat_response_json is False:
        print_err(chat_response, end="", error=False, date=False)
        print_err("chat response is not json")
        return multi_false

    bad_request = chat_response_json.get("bad_request", False)
    if bad_request is not False:
        print_err(bad_request, date=False, error=False)
        print_err("getting chat messages")
        return multi_false

    chat_response_file = robot_dir + "/chat-response"
    if not file_json_write(chat_response_file, chat_response_json):
        return multi_false

    peer_public_key = string_from_multiline_format(
        chat_response_json.get("peer_pubkey", False)
    )
    if peer_public_key is False:
        print_err("getting peer public key")
        return multi_false

    peer_fingerprint = gpg_import_key(peer_public_key)
    if peer_fingerprint is False:
        return multi_false

    if robot_save_peer_gpg_public_key(
        robot_dir, peer_public_key, peer_fingerprint, set_default=True
    ) is False:
        return multi_false

    return chat_response, chat_response_json


def message_get(message_date, message_nick, message_print, status_char, signature_error):
    return {
        "date": message_date,
        "nick": message_nick,
        "message": message_print,
        "status": status_char,
        "signature_error": signature_error
    }


def chat_print_single_message(message_dic):
    message_date = message_dic["date"]
    message_nick = message_dic["nick"]
    message_print = message_dic["message"]
    status_char = message_dic["status"]
    signature_error = message_dic["signature_error"]

    print_out(f"{message_date} {status_char} {message_nick}", end="")
    if signature_error is not None:
        print_out(f" {signature_error}", end="")
    print_out("\n", end="")
    print_out(message_print)

    return True


def chat_print_encrypted_messages(robot_dic, chat_response_json):
    robot_name, _, robot_dir, token, _, _, _ = robot_var_from_dic(robot_dic)

    unsorted_messages = chat_response_json.get("messages", False)
    if unsorted_messages is False:
        print_err(chat_response_json, end="", error=False, date=False)
        print_err("chat response does not have messages")
        return False

    messages = sorted(unsorted_messages, key=lambda mex: mex["time"])

    current_fingerprint = robot_get_current_fingerprint(robot_dir)
    if current_fingerprint is False:
        print_err("current key fingerprint is not present")
        return False

    peer_fingerprint = robot_get_peer_fingerprint(robot_dir, error_print=False)
    if peer_fingerprint is False:
        print_err("peer key fingerprint is not present")
        return False

    decrypted_messages = []

    first_message = True
    for message in messages:
        message_enc = string_from_multiline_format(message.get("message", False))
        if message_enc is False:
            print_err("getting encrypted message")
            return False

        message_time = message.get("time", False)
        if message_time is False:
            print_err("getting message time")
            return False

        message_nick = message.get("nick", False)
        if message_nick is False:
            print_err("getting message nick")
            return False

        message_date = date_convert_time_zone_and_format_string(message_time)

        signature_error = None
        if re.match('^#', message_enc) is not None:
            status_char = "N"
            message_print = message_enc
        elif token is not False:
            if message_nick == robot_name:
                signature_fingerprint = current_fingerprint
            else:
                signature_fingerprint = peer_fingerprint
            decrypted_message, signature_error = gpg_decrypt_check_message(
                message_enc, signature_fingerprint, passphrase=token, error_print=False
            )
            if decrypted_message is not False:
                message_print = decrypted_message
            else:
                message_print = "error decrypting message"
            if signature_error is None:
                status_char = "E"
            else:
                status_char = "X"
        else:
            print_err("token is False")
            return False

        if first_message:
            first_message = False
        else:
            print_out("\n", end="")

        message_dic = message_get(
            message_date, message_nick, message_print, status_char, signature_error
        )
        if message_dic is False:
            return False

        if chat_print_single_message(message_dic) is False:
            return False

        decrypted_messages.append(message_dic)

    decrypted_messages_file = robot_dir + "/messages-decrypted"
    if not file_json_write(decrypted_messages_file, decrypted_messages):
        return False

    return True


def robot_send_chat_message(robot_dic, message):
    robot_name, _, robot_dir, token, _, token_base91, robot_url = robot_var_from_dic(robot_dic)

    order_id = order_robot_get_last_order_id(robot_dic)
    if order_id is False:
        return False

    if re.match('^#', message) is not None:
        status_char = "N"
        message_send = message
    else:
        status_char = "E"
        current_fingerprint = robot_get_current_fingerprint(robot_dir)
        if current_fingerprint is False:
            return False

        peer_fingerprint = robot_get_peer_fingerprint(robot_dir, error_print=False)
        if peer_fingerprint is False:
            print_out("peer gpg key not yet present, searching it")
            chat_response, _ = robot_requests_chat(robot_dic)
            if chat_response is False:
                return False
            peer_fingerprint = robot_get_peer_fingerprint(robot_dir)
            if peer_fingerprint is False:
                return False

        message_send = string_to_multiline_format(gpg_encrypt_sign_message(
            message,
            [peer_fingerprint, current_fingerprint],
            current_fingerprint,
            passphrase=token
        ))
        if message_send is False:
            return False

    chat_response_post_all = requests_api_chat_post(
        token_base91, order_id, robot_url, robot_name, message_send
    )
    if response_is_error(chat_response_post_all):
        print_err(f"{robot_name} sending message")
        return False
    chat_response_post = chat_response_post_all.text
    chat_response_post_json = json_loads(chat_response_post)
    if chat_response_post_json is False:
        print_err(chat_response_post, end="\n", error=False, date=False)
        print_err("chat response is not json")
        return False

    print_out("message sent correctly")
    chat_print_single_message(message_get(
        "now", robot_name, message, status_char, None
    ))

    return True


def robot_send_chat_message_argv(argv: list):
    robot_dic, argv = robot_input_from_argv(argv)
    if robot_dic is False:
        return False

    if len(argv) < 1:
        message = input_ask("message: ")
    else:
        message = argv[0]
        argv = argv[1:]

    if message == "":
        print_err("message can not be empty")
        return False

    if not robot_send_chat_message(robot_dic, message):
        return False

    return True



#!/usr/bin/env python3

"""order.py"""

# pylint: disable=C0116 missing-function-docstring
# pylint: disable=C0209 consider-using-f-string

import os
import re

from roboauto.logger import print_out, print_err
from roboauto.global_state import roboauto_options, roboauto_state
from roboauto.order_data import \
    get_order_string, get_type_string, get_currency_string, \
    order_is_public, order_is_paused, order_is_waiting_maker_bond, \
    get_all_currencies, get_fiat_payment_methods, get_swap_payment_methods, \
    order_is_waiting_taker_bond, order_is_expired
from roboauto.order_local import \
    order_data_from_order_user, get_order_data, order_dic_from_robot_dir, \
    order_save_order_file, get_order_user, order_id_list_from_robot_dir, \
    order_is_this_hour_and_online, robot_set_make_response, \
    robot_order_set_local_make_data, robot_order_remove_local_make_data
from roboauto.robot import \
    robot_input_from_argv, robot_change_dir, robot_load_from_name, \
    robot_var_from_dic, robot_requests_get_order_id, robot_generate, \
    robot_list_dir
from roboauto.requests_api import \
    requests_api_order, requests_api_order_cancel, \
    requests_api_make, response_is_error, requests_api_order_take
from roboauto.utils import \
    json_dumps, string_is_false_none_null, file_is_executable, \
    json_loads, input_ask, roboauto_get_coordinator_from_url, \
    file_json_read, file_json_write, file_remove, bad_request_is_cancelled
from roboauto.subprocess_commands import subprocess_pay_invoice_and_check


def list_currencies():
    currencies = get_all_currencies()
    for _, currency in currencies.items():
        print_out(currency.lower())

    return True


def list_payment_methods(argv):
    fiat_present = True
    swap_present = True
    if len(argv) > 0:
        if argv[0] == "--fiat":
            swap_present = False
            argv = argv[1:]
        elif argv[0] == "--swap":
            fiat_present = False
            argv = argv[1:]

    payment_methods = []

    if fiat_present:
        payment_methods += get_fiat_payment_methods()

    if swap_present:
        payment_methods += get_swap_payment_methods()

    for method in payment_methods:
        print_out(method)

    return True


def order_user_empty_get(with_default):
    empty_order = get_order_user(
        False, False, False, False, False,
        False, False, False, False
    )

    if with_default:
        empty_order["public_duration"] = str(roboauto_options["default_duration"])
        empty_order["escrow_duration"] = str(roboauto_options["default_escrow"])
        empty_order["bond_size"] = str(roboauto_options["default_bond_size"])

    return empty_order


def list_order_fields():
    empty_order_user = order_user_empty_get(False)
    for item in empty_order_user:
        print_out(item)

    return True


def amount_correct_format(amount, is_fiat):
    if amount is False or amount is None or string_is_false_none_null(amount):
        return "null"

    if is_fiat is True:
        amount_format = "%.0f"
    else:
        amount_format = "%.3f"

    return amount_format % float(amount)


def peer_nick_from_response(order_response_json):
    if order_response_json.get("is_maker", True) is True:
        peer_nick = order_response_json.get("taker_nick", None)
    else:
        peer_nick = order_response_json.get("maker_nick", None)

    if string_is_false_none_null(peer_nick):
        return None

    return peer_nick


def order_requests_order_dic(
    robot_dic, order_id, order_function=None, take_amount=None,
    save_to_file=True, until_true=True, error_print_not_found_level=0, timeout=None
):
    """get the order_dic making a request to the coordinator
    order_function can be set to requests_api_order_take
    when taking an order
    will return False, or the string of bad_requests so it
    should be checked that the return value is not a string

    the order_dic is composed by 4 dictionaries:
    order_data:  can be derived from order_user, and is the data
                 used when creating orders
    order_urser: data user readable, everything that can be modivied
                 by the user
    order_info:  everything not in order_data and order_user
    order_response_json: the json response from the coordinator"""

    robot_name, _, robot_dir, _, _, token_base91, robot_url = robot_var_from_dic(robot_dic)

    if order_id is False or order_id is None:
        order_id = robot_requests_get_order_id(robot_dic)
        if order_id is False or order_id is None:
            return False

    requests_options = {
        "until_true": until_true,
        "error_print": error_print_not_found_level
    }
    if timeout is not None and timeout is not False:
        requests_options.update({
            "timeout": timeout
        })
    if order_function is None:
        order_response_all = requests_api_order(
            token_base91, order_id, robot_url, robot_name,
            options=requests_options
        )
    else:
        order_response_all = order_function(
            token_base91, order_id, robot_url, robot_name, take_amount=take_amount,
            options=requests_options
        )

    if response_is_error(order_response_all):
        print_err(f"{robot_name} {order_id} not found", level=error_print_not_found_level)
        return False
    order_response = order_response_all.text
    order_response_json = json_loads(order_response)
    if order_response_json is False:
        print_err(order_response, end="", error=False, date=False)
        print_err("getting order info for " + robot_name + " " + order_id)
        return False

    bad_request = order_response_json.get("bad_request", False)
    if bad_request is not False:
        print_err(bad_request, error=False, date=False)
        if not isinstance(bad_request, str):
            print_err("bad_request is not a string")
            print_err(f"{robot_name} {order_id} not available")
            return False
        else:
            if not bad_request_is_cancelled(bad_request):
                print_err(f"{robot_name} {order_id} not available")
            return bad_request

    coordinator = roboauto_get_coordinator_from_url(robot_url)

    status_id = order_response_json.get("status", False)
    type_id = order_response_json.get("type", False)
    currency_id = order_response_json.get("currency", False)
    amount = order_response_json.get("amount", False)
    has_range = order_response_json.get("has_range", False)
    min_amount = order_response_json.get("min_amount", False)
    max_amount = order_response_json.get("max_amount", False)
    payment_method = order_response_json.get("payment_method", False)
    premium = order_response_json.get("premium", False)
    public_duration = order_response_json.get(
        "public_duration", roboauto_options["default_duration"]
    )
    escrow_duration = order_response_json.get(
        "escrow_duration", roboauto_options["default_escrow"]
    )
    bond_size = order_response_json.get(
        "bond_size", str(roboauto_options["default_bond_size"])
    )
    is_taken = order_response_json.get("taker_locked", False)

    peer_nick = peer_nick_from_response(order_response_json)

    status_string = get_order_string(status_id)

    type_string = str(get_type_string(type_id))

    currency_string = str(get_currency_string(currency_id)).lower()
    is_fiat = currency_string != "btc"

    amount_single = None
    if is_taken or not has_range:
        amount_single = amount_correct_format(amount, is_fiat)
        if not amount_single:
            print_err(order_response, end="", error=False, date=False)
            print_err("format amount: " + amount)
            return False
    if not has_range:
        min_amount_user = amount
        max_amount_user = amount
        amount_string = amount_single
    else:
        min_amount_user = min_amount
        max_amount_user = max_amount
        min_amount_string = amount_correct_format(min_amount, is_fiat)
        max_amount_string = amount_correct_format(max_amount, is_fiat)
        if not min_amount_string or not max_amount_string:
            print_err(order_response, end="", error=False, date=False)
            print_err("format amount: " + min_amount + " " + max_amount)
            return False
        if is_taken:
            amount_string = amount_single
        else:
            amount_string = min_amount_string + "-" + max_amount_string

    order_description = \
        type_string + " " + currency_string + " " + str(amount_string) + " " + \
        premium + " " + payment_method + " " + status_string

    order_dic = {
        "order_data": get_order_data(
            type_id, currency_id,
            amount, has_range, min_amount, max_amount,
            payment_method, premium,
            public_duration, escrow_duration, bond_size
        ),
        "order_user": get_order_user(
            type_string, currency_string, min_amount_user, max_amount_user,
            payment_method, premium, str(public_duration), str(escrow_duration),
            bond_size
        ),
        "order_info": {
            "coordinator":          coordinator,
            "order_id":             order_id,
            "status":               status_id,
            "status_string":        status_string,
            "amount_string":        amount_string,
            "order_description":    order_description,
            "is_taken":             is_taken,
            "peer_nick":            peer_nick,
            "amount_single":        amount_single
        },
        "order_response_json":      order_response_json
    }

    if save_to_file is True:
        if not order_save_order_file(robot_dir, order_id, order_dic):
            return False

    return order_dic


# As a maker if you cancel an order after you have locked your maker bond,
# you are returned your bond. This may change in the future to prevent
# DDoSing the LN node and you won't be returned the maker bond.
def robot_cancel_order(robot_dic):
    robot_name, _, _, _, _, token_base91, robot_url = robot_var_from_dic(robot_dic)

    if robot_dic["state"] != "active":
        print_err("robot %s is not in the active directory" % robot_name)
        return False

    order_dic = order_requests_order_dic(robot_dic, order_id=False)
    if order_dic is False or isinstance(order_dic, str):
        return False

    order_info = order_dic["order_info"]

    order_id = order_info["order_id"]
    status_id = order_info["status"]

    if \
        not order_is_public(status_id) and \
        not order_is_paused(status_id) and \
        not order_is_waiting_maker_bond(status_id):
        print_err("robot order %s %s is not public or paused" % (robot_name, order_id))
        return False

    print_out("robot %s cancel order %s" % (robot_name, order_id))

    order_cancel_response_all = requests_api_order_cancel(
        token_base91, order_id, robot_url, robot_name
    )
    if response_is_error(order_cancel_response_all):
        return False
    order_cancel_response = order_cancel_response_all.text
    order_cancel_response_json = json_loads(order_cancel_response)
    if order_cancel_response_json is False:
        print_err(order_cancel_response, end="", error=False, date=False)
        print_err("cancelling order %s %s" % (robot_name, token_base91))
        return False

    bad_request = order_cancel_response_json.get("bad_request", False)
    if bad_request is False:
        print_err(order_cancel_response, end="", error=False, date=False)
        print_err("getting cancel response %s %s" % (robot_name, order_id))
        return False

    print_out(bad_request, date=False)

    return True


def amount_correct_from_response(order_response_json):
    amount_correct = order_response_json.get("amount", False)
    if amount_correct is False or amount_correct is None:
        return False

    is_fiat = str(get_currency_string(
        order_response_json.get("currency", "fiat")
    )).lower() != "btc"

    return amount_correct_format(amount_correct, is_fiat)


def premium_string_get(premium):
    if premium[0] == "-":
        return "below-" + premium[1:]
    else:
        return "above-" + premium


def order_string_status_print(robot_name, order_id, order_description, peer_nick):
    if peer_nick is False or peer_nick is None:
        print_out(f"{robot_name} {order_id} {order_description}")
    else:
        print_out(f"{robot_name} {peer_nick} {order_id} {order_description}")


def bond_order(robot_dic, order_id, taker=False, take_amount=None, use_node=True):
    """bond an order, after checking the invoice with lightning-node check
    will run lightning-node check and lightning-node pay as subprocesses
    if use_node is True it will simply print out the invoice"""

    robot_name = robot_dic["name"]

    if taker is False:
        order_function = None
    else:
        order_function = requests_api_order_take

    order_dic = order_requests_order_dic(
        robot_dic, order_id, order_function=order_function, take_amount=take_amount
    )
    if order_dic is False or isinstance(order_dic, str):
        return False

    # pylint: disable=R0801 duplicate-code
    order_user = order_dic["order_user"]
    order_info = order_dic["order_info"]
    order_response_json = order_dic["order_response_json"]

    order_description = order_info["order_description"]
    if taker is False:
        checking_function = order_is_waiting_maker_bond
        failure_function = order_is_expired
        if not checking_function(order_info["status"]):
            print_err(f"{robot_name} {order_id} is not waiting maker bond")
            return False
        string_paid = "bonded successfully, order is public"
        string_not_paid = "bond expired, order is not public"

        order_string_status_print(robot_name, order_id, order_description, None)

        name_pay_label = robot_name
    else:
        checking_function = order_is_waiting_taker_bond
        failure_function = order_is_public
        if not checking_function(order_info["status"]):
            print_err(f"order {order_id} is not waiting taker bond")
            return False
        string_paid = "bonded successfully, order is taken"
        string_not_paid = "bond expired, order is not taken"

        peer_nick = peer_nick_from_response(order_response_json)

        order_string_status_print(robot_name, order_id, order_description, peer_nick)

        name_pay_label = robot_name + "-" + str(peer_nick)

    bond_satoshis = order_response_json.get("bond_satoshis", False)
    if bond_satoshis is False:
        print_err(
            f"{robot_name} {order_id} bond_satoshis not present, invoice can not be checked"
        )
        return False

    bond_invoice = order_response_json.get("bond_invoice", False)
    if bond_invoice is False:
        print_err(
            f"{robot_name} {order_id} bond_invoice not present, invoice can not be checked"
        )
        return False

    if not use_node:
        print_out(bond_invoice)
        return True

    if not file_is_executable(roboauto_state["lightning_node_command"]):
        print_err("lightning node not set, to use without node pass --no-node")
        return False

    pay_label = \
        "bond-" + name_pay_label + "-" + order_id + "-" + \
        order_user["type"] + "-" + order_user["currency"] + "-" + \
        order_info["amount_string"] + "-" + premium_string_get(order_user["premium"])
    return subprocess_pay_invoice_and_check(
        robot_dic, order_id,
        bond_invoice, str(bond_satoshis), pay_label,
        lambda order_status : not checking_function(order_status),
        "checking if order is bonded...",
        string_paid, string_not_paid,
        order_requests_order_dic, failure_function
    )


def make_order(
    robot_dic, make_data, should_bond=True, check_change=False, use_node=True
):
    """make the request to the coordinator to create an order.
    if should_bond is true, also bond it, if it is false do not create
    the order if there are more than maximum order this hour, save
    make data to file instead
    if check_change is true, change the order from data saved on disk
    return False if an error None if the requests was not made"""

    robot_name, robot_state, robot_dir, _, coordinator, token_base91, robot_url = \
        robot_var_from_dic(robot_dic)

    change_order_file = robot_dir + "/change-next-order"

    if check_change is True:
        if os.path.isfile(change_order_file):
            change_order = file_json_read(change_order_file)
            if change_order is not False:
                for key, value in change_order.items():
                    if value is not False:
                        old_value = make_data[key]
                        make_data[key] = value
                        print_out(
                            f"{robot_name} {key} "
                            f"changed from {old_value} to {value}"
                        )

    if roboauto_options["robot_maximum_orders"] > 0:
        max_orders = roboauto_options["robot_maximum_orders"]
        order_ids = order_id_list_from_robot_dir(robot_dir, error_print=False)
        if \
            order_ids is not False and order_ids is not None and \
            len(order_ids) >= max_orders:
            print_out(
                f"{robot_name} {coordinator} maximum orders {max_orders} reached"
            )
            if roboauto_options["create_new_after_maximum_orders"] is True:
                new_robot_dic = robot_generate(coordinator, robot_state)
                if new_robot_dic is False:
                    return False
                new_robot_name = new_robot_dic["name"]
                new_robot_dir = new_robot_dic["dir"]

                print_out(f"{new_robot_name} created with order data of {robot_name}")

                if robot_order_set_local_make_data(new_robot_dir, make_data) is False:
                    return False

                print_out(f"{robot_name} moving to inactive")
                if not robot_change_dir(robot_name, "inactive"):
                    print_err(f"{robot_name} moving to inactive")
                    return False

                return make_order(
                    new_robot_dic, make_data,
                    should_bond=should_bond,
                    check_change=check_change,
                    use_node=use_node
                )
            else:
                print_out(f"{robot_name} moving to inactive")
                if not robot_change_dir(robot_name, "inactive"):
                    print_err(f"{robot_name} moving to inactive")
                    return False

                return None

    if should_bond is False:
        robot_this_hour = 0

        active_list = robot_list_dir(roboauto_state["active_home"])
        for robot_active in active_list:
            # pylint: disable=R0801 duplicate-code
            order_dic = order_dic_from_robot_dir(
                roboauto_state["active_home"] + "/" + robot_active,
                order_id=None, error_print=False
            )
            if order_dic is False or order_dic is None:
                continue

            if order_is_this_hour_and_online(order_dic):
                robot_this_hour += 1

        order_maximum = roboauto_options["order_maximum"]
        if robot_this_hour >= order_maximum:
            print_out(
                f"{robot_name} there are already {order_maximum} orders this hour, " +
                "saving make data to file without creating order"
            )
            return robot_order_set_local_make_data(robot_dir, make_data)

    if not make_data["bond_size"]:
        print_err("bond size percentage not defined")
        return False

    make_response_all = requests_api_make(
        token_base91, robot_url, robot_name, make_data=json_dumps(make_data)
    )
    if response_is_error(make_response_all):
        return False
    make_response = make_response_all.text
    make_response_json = json_loads(make_response)
    if make_response_json is False:
        print_err(make_data, error=False, date=False)
        print_err("getting response make order for " + robot_name)
        return False
    order_id_number = make_response_json.get("id", False)
    if order_id_number is False:
        bad_request = make_response_json.get("bad_request", False)
        if bad_request is not False:
            print_err(bad_request, error=False, date=False)
            print_err("making order")
            if "Your order maximum amount is too big" in bad_request:
                if not robot_change_dir(robot_name, "paused", error_is_already=False):
                    return False
        else:
            print_err(make_response, end="", error=False, date=False)
            print_err(make_data, error=False, date=False)
            print_err("getting id of new order for " + robot_name)
        return False

    if not robot_set_make_response(robot_dir, make_response_json):
        return False
    if not robot_order_remove_local_make_data(robot_dir):
        return False

    order_id = str(order_id_number)

    if check_change is True:
        if os.path.isfile(change_order_file):
            file_remove(change_order_file)

    if not should_bond:
        print_out(f"{robot_name} order {order_id} will not be bonded")
        return True

    return bond_order(robot_dic, order_id, use_node=use_node)


def order_user_from_argv(argv, with_default=False, only_set=False):
    """get order_user's fields from argv"""
    order_user = order_user_empty_get(with_default)

    if len(argv) > 0 and argv[0] == "--from-robot":
        argv = argv[1:]
        if len(argv) < 1:
            print_err("insert robot")
            return False
        robot_name = argv[0]
        argv = argv[1:]

        robot_dic = robot_load_from_name(robot_name)
        if robot_dic is False:
            return False

        order_dic = order_dic_from_robot_dir(robot_dic["dir"])
        if order_dic is False or order_dic is None:
            return False

        order_user = order_dic.get("order_user", False)
        if order_user is False:
            print_err(f"{robot_name} does not have order user")
            return False
    else:
        while len(argv) > 0:
            param = argv[0]
            argv = argv[1:]
            key_value = param.split("=", 1)
            if len(key_value) != 2:
                print_err("parameter %s is not key=value" % param)
                return False
            key, value = key_value
            if key in order_user:
                if order_user[key] is False:
                    order_user[key] = value
                else:
                    if key != "payment_method":
                        print_err(f"{key} can not be set multiple times")
                        return False
                    else:
                        order_user[key] += " " + value
            else:
                print_err("%s is not a valid key" % key)
                return False

    if only_set is False:
        return order_user
    else:
        return {k: v for k, v in order_user.items() if v is not False}


def create_order(argv):
    should_bond = True
    use_node = True
    should_set_active = True
    while len(argv) > 0:
        if argv[0] == "--no-bond":
            should_bond = False
            argv = argv[1:]
        elif argv[0] == "--no-node":
            use_node = False
            argv = argv[1:]
        elif argv[0] == "--no-active":
            should_set_active = False
            argv = argv[1:]
        elif re.match('^-', argv[0]) is not None:
            print_err("option %s not recognized" % argv[0])
            return False
        else:
            break

    if not should_bond and not use_node:
        print_err("--no-bond and --no-node should not both be present")
        return False

    # pylint: disable=R0801 duplicate-code
    robot_dic, argv = robot_input_from_argv(argv)
    if robot_dic is False:
        return False

    robot_name = robot_dic["name"]

    if robot_dic["state"] != "paused":
        print_err("robot %s is not in the paused directory" % robot_name)
        return False

    order_user = order_user_from_argv(argv, with_default=True)
    if order_user is False:
        return False

    for key, value in order_user.items():
        if value is False:
            answer = input_ask("-%24s" % (key + ":"))
            if answer is False:
                return False
            order_user[key] = answer

    order_data = order_data_from_order_user(order_user)
    if order_data is False:
        return False

    make_order_result = make_order(
        robot_dic,
        order_data,
        should_bond=should_bond,
        use_node=use_node
    )
    if make_order_result is False or make_order_result is None:
        return False

    if use_node:
        print_out(f"{robot_name} order created successfully")

    if should_set_active is True:
        if not robot_change_dir(robot_name, "active"):
            return False

    return True


def cancel_order(argv):
    robot_dic, argv = robot_input_from_argv(argv)
    if robot_dic is False:
        return False

    robot_name = robot_dic["name"]

    if robot_dic["state"] != "active":
        print_err("robot %s is not in the active directory" % robot_name)
        return False

    if not robot_cancel_order(robot_dic):
        return False

    if not robot_change_dir(robot_name, "inactive"):
        return False

    print_out("%s moved to inactive" % robot_name)

    return True


def recreate_order(argv):
    should_cancel = True
    use_node = True
    should_bond = True
    while len(argv) >= 1:
        if argv[0] == "--no-bond":
            should_bond = False
            argv = argv[1:]
        if argv[0] == "--no-node":
            use_node = False
            argv = argv[1:]
        elif argv[0] == "--no-cancel":
            should_cancel = False
            argv = argv[1:]
        elif re.match('^-', argv[0]) is not None:
            print_err("option %s not recognized" % argv[0])
            return False
        else:
            break

    if not should_bond and not use_node:
        print_err("--no-bond and --no-node should not both be present")
        return False

    # pylint: disable=R0801 duplicate-code
    robot_dic, argv = robot_input_from_argv(argv)
    if robot_dic is False:
        return False

    robot_name = robot_dic["name"]
    robot_dir = robot_dic["dir"]

    if should_cancel:
        if robot_dic["state"] != "active":
            print_err("robot %s is not in the active directory" % robot_name)
            return False
    else:
        if robot_dic["state"] not in ("paused", "inactive"):
            print_err("robot %s is not in the paused or inactive directories" % robot_name)
            return False

    order_user = order_user_from_argv(argv)
    if order_user is False:
        return False

    if should_cancel:
        if not robot_cancel_order(robot_dic):
            return False

    order_dic = order_dic_from_robot_dir(robot_dir, order_id=None)
    if order_dic is False or order_dic is None:
        return False

    order_user_old = order_dic["order_user"]
    for key, value in order_user.items():
        if value is False:
            order_user[key] = order_user_old[key]

    order_data = order_data_from_order_user(order_user)
    if order_data is False:
        return False

    make_order_result = make_order(
        robot_dic,
        order_data,
        should_bond=should_bond,
        use_node=use_node
    )
    if make_order_result is False or make_order_result is None:
        return False

    if use_node:
        print_out(f"{robot_name} order recreated successfully")

    if should_cancel is False:
        if not robot_change_dir(robot_name, "active"):
            return False

    return True


def order_change_next_expire(argv):
    action = "set"
    if len(argv) > 0:
        first_arg = argv[0]
        if first_arg in ("--print", "--remove"):
            argv = argv[1:]
            action = first_arg.split("-", 2)[2]

    robot_dic, argv = robot_input_from_argv(argv)
    if robot_dic is False:
        return False

    robot_name, _, robot_dir, _, _, _, _ = robot_var_from_dic(robot_dic)

    change_order_file = robot_dir + "/change-next-order"

    if action in ("print", "remove"):
        if not os.path.isfile(change_order_file):
            print_out(f"{robot_name} change order not set")
        else:
            change_order = file_json_read(change_order_file)
            if change_order is False:
                return False
            print_out(json_dumps(change_order))

            if action == "remove":
                if not file_remove(change_order_file):
                    return False
    elif action == "set":
        order_user = order_user_from_argv(argv, only_set=True)
        if order_user is False:
            return False

        something_set = False
        for _, value in order_user.items():
            if value is not False:
                something_set = True
                break
        if something_set is False:
            print_err("nothing set to change")
            return False

        if not file_json_write(change_order_file, order_user):
            return False

        print_out(json_dumps(order_user))

    return True



#!/usr/bin/env python3

"""keep_online.py"""

# pylint: disable=C0116 missing-function-docstring

import time

import filelock

from roboauto.logger import print_out, print_err, logger_flush
from roboauto.global_state import roboauto_state, roboauto_options
from roboauto.robot import \
    robot_list_dir, robot_load_from_name, waiting_queue_get, \
    robot_change_dir, robot_get_dir_dic, robot_wait, \
    robot_unwait, robot_check_and_claim_reward, \
    robot_requests_get_order_id
from roboauto.order_data import  \
    order_is_public, order_is_paused, order_is_finished, \
    order_is_pending, order_is_waiting_maker_bond, \
    order_is_waiting_taker_bond, order_is_expired, \
    order_is_finished_for_seller, order_is_waiting_seller_buyer, \
    order_is_waiting_seller, order_is_waiting_buyer, \
    order_is_failed_routing, get_order_expiry_reason_string, \
    order_expired_is_not_taken, order_expired_is_maker_bond_not_locked
from roboauto.order_local import \
    robot_handle_taken, order_dic_from_robot_dir, \
    order_robot_get_last_order_id, order_save_order_file, \
    order_is_this_hour_and_online, robot_have_make_data, \
    robot_order_get_local_make_data
from roboauto.order import \
    order_requests_order_dic, bond_order, make_order, \
    peer_nick_from_response
from roboauto.order_action import \
    order_seller_bond_escrow, order_buyer_update_invoice
from roboauto.date_utils import \
    get_current_timestamp, get_current_minutes_from_timestamp, \
    timestamp_from_date_string, date_convert_time_zone_and_format_string
from roboauto.utils import \
    update_roboauto_options, lock_file_name_get, \
    shuffle_dic, file_is_executable, arg_key_value_number, \
    bad_request_is_cancelled, bad_request_is_wrong_robot


def robot_handle_active_expired(robot_dic, all_dic, make_data, expiry_reason=None):
    robot_name = robot_dic["name"]
    robot_dir = robot_dic["dir"]

    if expiry_reason is not None and expiry_reason is not False:
        if \
            not order_expired_is_not_taken(expiry_reason) and \
            not order_expired_is_maker_bond_not_locked(expiry_reason):
            expiry_string = get_order_expiry_reason_string(expiry_reason)
            print_out(f"{robot_name} was active and now is {expiry_string}")

            earned_rewards = robot_check_and_claim_reward(robot_dic)
            if earned_rewards is False:
                return False
            elif earned_rewards > 0:
                # while there are rewards to be claimed it is not moving from active
                return True
            print_out(f"{robot_name} unusual expiry, moving to paused")
            return robot_change_dir(robot_name, "paused")

    if count_active_orders_this_hour(all_dic) < roboauto_options["order_maximum"]:
        if make_data is None or make_data is False:
            make_data = robot_order_get_local_make_data(robot_dir)
            if make_data is None or make_data is False:
                print_out(
                    f"{robot_name} does not have orders and make data, " +
                    "moving to paused"
                )
                return robot_change_dir(robot_name, "paused")
            print_out(f"{robot_name} creating order from make data")

        # will return None when maxium robot orders is reached,
        # it is ok to return True since the robot is checked correctly
        if make_order(
            robot_dic,
            make_data,
            check_change=True
        ) is False:
            return False
    else:
        if not robot_wait(robot_name):
            return False

    return True


def robot_handle_active(robot_dic, all_dic):
    """handle an active robot"""

    robot_name = robot_dic["name"]
    robot_dir = robot_dic["dir"]
    robot_coordinator = robot_dic["coordinator"]

    # handle robot with make data
    if robot_have_make_data(robot_dir):
        return robot_handle_active_expired(robot_dic, all_dic, None)

    order_id = order_robot_get_last_order_id(robot_dic, error_print=False)
    if order_id is False or order_id is None:
        print_out(f"{robot_name} active does not have orders saved, making request")
        order_id = robot_requests_get_order_id(robot_dic, error_print_not_found_level=2)
        if order_id is False or order_id is None:
            return False

    old_order_dic = order_dic_from_robot_dir(robot_dir, order_id, error_print=False)

    # save to file just when status id is different from previous
    order_dic = order_requests_order_dic(
        robot_dic, order_id, save_to_file=False,
        until_true=False, error_print_not_found_level=2,
        timeout=roboauto_options["orders_timeout"]
    )
    if order_dic is False:
        return False
    elif bad_request_is_wrong_robot(order_dic):
        print_out(f"{robot_name} {order_id} wrong robot, moving to paused")
        return robot_change_dir(robot_name, "paused")
    elif bad_request_is_cancelled(order_dic):
        earned_rewards = robot_check_and_claim_reward(robot_dic)
        if earned_rewards is False:
            return False
        elif earned_rewards > 0:
            return True
        print_out(f"{robot_name} {order_id} active is cancelled, moving to inactive")
        return robot_change_dir(robot_name, "inactive")
    elif isinstance(order_dic, str):
        return False

    order_info = order_dic["order_info"]

    status_id = order_info["status"]

    # do not save when is waiting taken bond because it will change expires_at
    # and the robot will not appear to be online
    if \
        old_order_dic is False or old_order_dic is None or (
            status_id != old_order_dic["order_info"]["status"] and
            not order_is_waiting_taker_bond(status_id)
        ):
        if not order_save_order_file(robot_dir, order_id, order_dic):
            return False

    nicks_waiting = waiting_queue_get()
    if robot_name in nicks_waiting:
        if order_is_public(status_id):
            if robot_unwait(nicks_waiting, robot_name) is False:
                return False
            print_out(f"{robot_name} removed from waiting queue because it is active")
        if \
            not order_is_waiting_maker_bond(status_id) and \
            not order_is_public(status_id) and \
            not order_is_paused(status_id) and \
            not order_is_waiting_taker_bond(status_id) and \
            not order_is_expired(status_id):
            if robot_unwait(nicks_waiting, robot_name) is False:
                return False
            print_out(f"{robot_name} removed from waiting queue because it is taken")
            if not robot_handle_taken(
                robot_name, status_id, order_id, order_info["order_description"]
            ):
                return False
    else:
        if order_is_public(status_id):
            return True

        status_string = order_info["status_string"]
        print_out(f"{robot_name} {robot_coordinator} {order_id} {status_string}")

        if order_is_paused(status_id):
            print_out(robot_name + " " + order_id + " " + order_info["order_description"])
            print_out(robot_name + " " + order_id + " moving to paused")
            if not robot_change_dir(robot_name, "paused"):
                print_err("moving " + robot_name + " to paused")
                return False
        elif order_is_waiting_taker_bond(status_id):
            print_out(robot_name + " " + order_id + " is in the process of being taken")
        elif order_is_waiting_maker_bond(status_id):
            if count_active_orders_this_hour(all_dic) < roboauto_options["order_maximum"]:
                if bond_order(robot_dic, order_id) is False:
                    return False
            else:
                if not robot_wait(robot_name):
                    return False
        elif order_is_expired(status_id):
            return robot_handle_active_expired(
                robot_dic, all_dic, order_dic["order_data"],
                expiry_reason=order_dic["order_response_json"].get("expiry_reason", None)
            )
        else:
            if not robot_handle_taken(
                robot_name, status_id, order_id, order_info["order_description"]
            ):
                return False

    return True


def count_active_orders_this_hour(all_dic):
    robot_this_hour = 0

    for robot_name, robot_info in all_dic.items():
        robot_state = robot_info["state"]
        if robot_state != "active":
            continue

        order_dic = order_dic_from_robot_dir(
            roboauto_state["active_home"] + "/" + robot_name,
            order_id=None, error_print=False
        )
        if order_dic is False or order_dic is None:
            continue

        if order_is_this_hour_and_online(order_dic):
            robot_this_hour += 1

    return robot_this_hour


def should_remove_from_waiting_queue(all_dic):
    current_timestamp = get_current_timestamp()

    # let old orders expire
    current_minutes = get_current_minutes_from_timestamp(current_timestamp)
    min_minutes = roboauto_state["waiting_queue_remove_after"]
    if \
        current_minutes > min_minutes and \
        count_active_orders_this_hour(all_dic) < roboauto_options["order_maximum"] and \
        len(waiting_queue_get()) > 0:
        return True

    return False


def pending_robot_should_act(expires_timestamp, escrow_duration):
    remaining_seconds = expires_timestamp - get_current_timestamp()
    if remaining_seconds < 0:
        return False

    seconds_pending_order = roboauto_options["seconds_pending_order"]
    if seconds_pending_order > 0:
        if remaining_seconds < seconds_pending_order:
            return True
    elif seconds_pending_order < 0:
        if remaining_seconds < escrow_duration + seconds_pending_order:
            return True

    return False


def robot_handle_pending(robot_dic):
    robot_name = robot_dic["name"]
    robot_dir = robot_dic["dir"]
    robot_coordinator = robot_dic["coordinator"]

    order_id = order_robot_get_last_order_id(robot_dic, error_print=False)
    if order_id is False or order_id is None:
        print_out(f"{robot_name} pending does not have orders saved, making request")
        order_id = robot_requests_get_order_id(robot_dic, error_print_not_found_level=2)
        if order_id is False or order_id is None:
            return False

    old_order_dic = order_dic_from_robot_dir(robot_dir, order_id, error_print=False)

    # save to file just when status id is different from previous
    order_dic = order_requests_order_dic(
        robot_dic, order_id, save_to_file=False,
        until_true=False, error_print_not_found_level=2,
        timeout=roboauto_options["orders_timeout"]
    )
    if order_dic is False:
        return False
    elif bad_request_is_wrong_robot(order_dic):
        print_out(f"{robot_name} {order_id} wrong robot, moving to paused")
        return robot_change_dir(robot_name, "paused")
    elif bad_request_is_cancelled(order_dic):
        earned_rewards = robot_check_and_claim_reward(robot_dic)
        if earned_rewards is False:
            return False
        elif earned_rewards > 0:
            return True
        print_out(f"{robot_name} {order_id} pending is cancelled, moving to inactive")
        return robot_change_dir(robot_name, "inactive")
    elif isinstance(order_dic, str):
        return False

    order_response_json = order_dic["order_response_json"]
    order_info = order_dic["order_info"]

    status_id = order_info["status"]

    is_seller = order_response_json.get("is_seller", False)

    if \
        old_order_dic is False or old_order_dic is None or \
        status_id != old_order_dic["order_info"]["status"]:
        if not order_save_order_file(robot_dir, order_id, order_dic):
            return False

    if \
        order_is_pending(status_id) and \
        not (is_seller and order_is_finished_for_seller(status_id)):
        if not is_seller and order_is_failed_routing(status_id):
            if "trade_satoshis" in order_response_json:
                print_out(
                    f"{robot_name} {order_id} old invoice failed, sending a new one"
                )
                return order_buyer_update_invoice(robot_dic, (None, None))
            else:
                order_description = order_info["order_description"]
                peer_nick = peer_nick_from_response(order_response_json)
                print_out(f"{robot_name} {peer_nick} {order_id} {order_description}")
                if "failure_reason" in order_response_json:
                    failure_reason = order_response_json["failure_reason"]
                    print_out(f"{robot_name} {peer_nick} {order_id} {failure_reason}")
                return True

        if \
            order_is_waiting_seller_buyer(status_id) or \
            order_is_waiting_seller(status_id) or \
            order_is_waiting_buyer(status_id):
            expires_at = order_response_json.get("expires_at", False)
            if expires_at is False:
                print_err("no expires_at")
                return False

            escrow_duration = order_response_json.get("escrow_duration", False)
            if escrow_duration is False:
                print_err("no escrow_duration")
                return False

            if pending_robot_should_act(
                timestamp_from_date_string(expires_at),
                int(escrow_duration)
            ):
                date_short_expire = date_convert_time_zone_and_format_string(
                    expires_at, output_format="%H:%M:%S"
                )
                if is_seller:
                    if \
                        order_is_waiting_seller_buyer(status_id) or \
                        order_is_waiting_seller(status_id):
                        print_out(
                            f"{robot_name} {order_id} "
                            f"expires at {date_short_expire}, paying escrow"
                        )
                        return order_seller_bond_escrow(robot_dic, True)
                else:
                    if \
                        order_is_waiting_seller_buyer(status_id) or \
                        order_is_waiting_buyer(status_id):
                        print_out(
                            f"{robot_name} {order_id} "
                            f"expires at {date_short_expire}, sending invoice"
                        )
                        return order_buyer_update_invoice(robot_dic, (None, None))
    else:
        status_string = order_info["status_string"]
        print_out(f"{robot_name} {robot_coordinator} {order_id} {status_string}")

        earned_rewards = robot_check_and_claim_reward(robot_dic)
        if earned_rewards is False:
            return False
        elif earned_rewards > 0:
            # while there are rewards to be claimed it is not moving from pending
            return True

        expiry_reason = order_response_json.get("expiry_reason", -1)

        if \
            order_is_finished(status_id) or \
            (is_seller and order_is_finished_for_seller(status_id)):
            print_out(f"{robot_name} {order_id} is completed, moving to inactive")
            return robot_change_dir(robot_name, "inactive")
        elif order_is_public(status_id):
            print_out(
                f"{robot_name} {order_id} was pending and now is public, " +
                "moving to active"
            )
            return robot_change_dir(robot_name, "active")
        elif order_is_expired(status_id):
            if order_expired_is_not_taken(expiry_reason):
                print_out(
                    f"{robot_name} {order_id} was pending and now is expired not taken, " +
                    "moving to active"
                )
                return robot_change_dir(robot_name, "active")
            else:
                expiry_string = get_order_expiry_reason_string(expiry_reason)
                print_out(
                    f"{robot_name} {order_id} was pending and now is {expiry_string}, " +
                    "moving to paused"
                )
                return robot_change_dir(robot_name, "paused")
        else:
            print_err(f"{robot_name} strange state, moving to paused")
            return robot_change_dir(robot_name, "paused")

    return True


def robot_check_last_checked(robot_dic, seconds_not_checked):
    robot_name = robot_dic["name"]
    robot_dir = robot_dic["dir"]

    order_dic = order_dic_from_robot_dir(
        robot_dir, error_print=False
    )
    if order_dic is False or order_dic is None:
        return False

    order_data = order_dic.get("order_data", False)
    if order_data is False:
        return False
    escrow_duration = order_data.get("escrow_duration", False)
    if not isinstance(escrow_duration, int):
        return False

    if seconds_not_checked > int(escrow_duration / 2):
        minutes_not_chcked = int(seconds_not_checked / 60)
        minutes_escrow = int (escrow_duration / 60)
        print_out(
            f"{robot_name} was not successfully checked " +
            f"for {minutes_not_chcked} minutes, " +
            f"escrow duration {minutes_escrow} minutes"
        )

    return True


def robot_dic_update(all_dic):
    """return number of added and failed robots"""

    active_set = robot_list_dir(robot_get_dir_dic()["active"], get_set=True)
    pending_set = robot_list_dir(robot_get_dir_dic()["pending"], get_set=True)

    added_robots = 0
    failed_robots = 0

    for robot_name, robot_info in list(all_dic.items()):
        robot_state = robot_info["state"]
        if robot_state == "active":
            if robot_name not in active_set:
                nicks_waiting = waiting_queue_get()
                if robot_name in nicks_waiting:
                    if robot_unwait(nicks_waiting, robot_name) is not False:
                        print_out(\
                            f"{robot_name} removed from waiting queue " +
                            "because it is no longer active"
                        )
                if robot_name in pending_set:
                    print_out(f"{robot_name} moved from active to pending")
                    all_dic[robot_name]["state"] = "pending"
                else:
                    all_dic.pop(robot_name)
                    print_out(f"{robot_name} removed from active directory")
        elif robot_state == "pending":
            if robot_name not in pending_set:
                if robot_name in active_set:
                    print_out(f"{robot_name} moved from pending to active")
                    all_dic[robot_name]["state"] = "active"
                else:
                    all_dic.pop(robot_name)
                    print_out(f"{robot_name} removed from pending directory")

    for robot_active in active_set:
        if robot_active not in all_dic:
            print_out(f"{robot_active} added to active directory")

            if count_active_orders_this_hour(all_dic) >= roboauto_options["order_maximum"]:
                robot_wait(robot_active)

            all_dic[robot_active] = {
                "state": "active",
                "last_checked": int(time.time())
            }

            added_robots += 1

            robot_dic = robot_load_from_name(robot_active)
            if robot_dic is False:
                print_err(f"{robot_active} getting active robot")
                failed_robots += 1
                continue

            if robot_handle_active(robot_dic, all_dic) is False:
                failed_robots += 1

    for robot_pending in pending_set:
        if robot_pending not in all_dic:
            print_out(f"{robot_pending} added to pending directory")

            all_dic[robot_pending] = {
                "state": "pending",
                "last_checked": int(time.time())
            }

            added_robots += 1

            robot_dic = robot_load_from_name(robot_pending)
            if robot_dic is False:
                print_err(f"{robot_pending} getting pending robot")
                failed_robots +=1
                continue

            if robot_handle_pending(robot_dic) is False:
                failed_robots += 1

    return added_robots, failed_robots


# active_dic is the set of current active robots
# with an associated time % roboauto_options["active_interval"]
# pending_dic is the set of current pending robots
# with an associated time % roboauto_options["pending_interval"]
# this way robots are not checked all together,
# but every one is checked at a different time
def keep_online_no_lock(should_sleep, initial_info):
    active_list = robot_list_dir(roboauto_state["active_home"])
    pending_list = robot_list_dir(roboauto_state["pending_home"])

    if len(active_list) < 1 and len(pending_list) < 1:
        print_out("there are no active or pending robots", date=False)
        return True

    if initial_info:
        if len(active_list) >= 1:
            print_out(f"current {len(active_list)} active robots are:", date=False)
            for robot_name in active_list:
                print_out(robot_name, date=False)
        else:
            print_out("there are currently no active robots", date=False)
        print_out("\n", end="", date=False)

        if len(pending_list) >= 1:
            print_out(f"current {len(pending_list)} pending robots are:", date=False)
            for robot_name in pending_list:
                print_out(robot_name, date=False)
        else:
            print_out("there are currently no pending robots", date=False)
        print_out("\n", end="", date=False)

    ordered_all_dic = {}

    all_starting_time = time.time()

    for active_robot in active_list:
        ordered_all_dic[active_robot] = {
            "state": "active",
            "last_checked": int(all_starting_time)
        }

    for pending_robot in pending_list:
        ordered_all_dic[pending_robot] = {
            "state": "pending",
            "last_checked": int(all_starting_time)
        }

    all_dic = shuffle_dic(ordered_all_dic)

    robot_check_current = 0

    logger_flush()

    while True:
        all_starting_time = time.time()

        total_robots = len(all_dic)
        if robot_check_current >= total_robots:
            robot_check_current = 0

        if len(all_dic) < 1:
            print_out("there are no active or pending robots", date=False)
            return True

        # every loop make a robot requests to check for rewards
        robot_name = list(all_dic.keys())[robot_check_current]
        robot_dic = robot_load_from_name(robot_name)
        if robot_dic is False:
            print_err(f"{robot_name} skipping request robot")
        else:
            _ = robot_check_and_claim_reward(robot_dic, error_print_not_found_level=2)

        robot_check_current += 1

        failed_numbers = 0

        for robot_name, robot_info in list(all_dic.items()):
            if robot_name not in all_dic:
                total_robots -= 1
                continue
            robot_state = robot_info["state"]

            starting_time = time.time()

            robot_dic = robot_load_from_name(robot_name)
            if robot_dic is False:
                print_err(f"{robot_name} skipping {robot_state} robot")
                failed_numbers += 1
                continue

            if robot_state == "active":
                if robot_handle_active(robot_dic, all_dic) is False:
                    failed_numbers += 1
                    robot_check_last_checked(
                        robot_dic, int(starting_time) - robot_info["last_checked"]
                    )
                else:
                    all_dic[robot_name]["last_checked"] = int(starting_time)
            elif robot_state == "pending":
                if robot_handle_pending(robot_dic) is False:
                    failed_numbers += 1
                    robot_check_last_checked(
                        robot_dic, int(starting_time) - robot_info["last_checked"]
                    )
                else:
                    all_dic[robot_name]["last_checked"] = int(starting_time)

            added_robots, additional_failed_robots = robot_dic_update(all_dic)
            total_robots += added_robots
            failed_numbers += additional_failed_robots

            # allow to adjust configs while roboauto is running
            if update_roboauto_options(True) is False:
                print_err("reading the config file")

            if len(all_dic) < 1:
                print_out("there are no active or pending robots", date=False)
                return True

            logger_flush()

            elapsed_time = time.time() - starting_time

            # robosats/api/logics.py user_activity_status
            # 2 minutes Active
            # 10 minutes Seen recently

            if should_sleep is True:
                half_max_time = (120 / len(all_dic)) / 2
                if elapsed_time < half_max_time:
                    time.sleep(half_max_time)

        if should_remove_from_waiting_queue(all_dic):
            robot_unwaited = robot_unwait(waiting_queue_get())
            if robot_unwaited is not False:
                print_out(f"{robot_unwaited} removed from waiting queue")

        all_elapsed_time = int(time.time() - all_starting_time)
        if failed_numbers < total_robots / 2:
            print_out(
                f"{total_robots} robots checked in {all_elapsed_time} seconds " +
                f"{failed_numbers} failed",
                level=1
            )
        else:
            print_out(
                f"{total_robots} robots checked in {all_elapsed_time} seconds " +
                f"{failed_numbers} failed, connection may be instable",
                level=0
            )
        logger_flush()


def keep_online(argv):
    should_sleep = True
    initial_info = True
    while len(argv) >= 1:
        current_arg = argv[0]
        argv = argv[1:]

        if current_arg == "--no-sleep":
            should_sleep = False
        elif current_arg == "--no-initial-info":
            initial_info = False
        else:
            arg_verbosity = arg_key_value_number("verbosity", current_arg)
            if arg_verbosity is False:
                return False
            elif arg_verbosity is not None:
                roboauto_state["log_level"] = arg_verbosity
            else:
                print_err(f"option {current_arg} not recognied", date=False, error=False)
                return False

    if not file_is_executable(roboauto_state["lightning_node_command"]):
        print_err("lightning node not set, it is required for keep-online")
        return False

    try:
        with filelock.FileLock(
            lock_file_name_get("keep-online"),
            timeout=0
        ):
            return keep_online_no_lock(should_sleep, initial_info)
    except filelock.Timeout:
        print_err("keep online is already running", date=False, error=False)

    return False



#!/usr/bin/env python3

"""global_state.py"""

# options that can be changed in config file
roboauto_options = {
    "user_agent": "Mozilla/5.0 (Windows NT 10.0; rv:109.0) Gecko/20100101 Firefox/115.0",
    "date_format": "%Y/%m/%d %H:%M:%S",
    "seconds_pending_order": 0,
    "order_maximum": 2,
    "robot_maximum_orders": 0,
    "create_new_after_maximum_orders": False,
    "tab_size": 4,
    "routing_budget_ppm": 1000,
    "requests_timeout": 120,
    "orders_timeout": 60,
    "active_interval": 80,
    "pending_interval": 120,
    "pay_interval": 2,
    "error_interval": 5,
    "default_duration": 86400,
    "default_escrow": 28800,
    "default_bond_size": 3.00,
    "federation": {
        "exp": None,
        # "exp": "http://robosats6tkf3eva7x2voqso3a5wcorsnw34jveyxfqi2fu7oyheasid.onion",
        "sau": "http://satstraoq35jffvkgpfoqld32nzw2siuvowanruindbfojowpwsjdgad.onion",
        "tos": "http://ngdk7ocdzmz5kzsysa3om6du7ycj2evxp2f2olfkyq37htx3gllwp2yd.onion",
        "tbl": "http://4t4jxmivv6uqej6xzx2jx3fxh75gtt65v3szjoqmc4ugdlhipzdat6yd.onion",
        "bve": "http://mmhaqzuirth5rx7gl24d4773lknltjhik57k7ahec5iefktezv4b3uid.onion"
    }
}

# program state and options that can not be changed in the config file
roboauto_state = {
    "current_command_type": "info",
    "keep_online_hour_relative": False,
    "gpg": None,
    "logger": None,
    "log_level": 0,
    "filelock_timeout": 120,
    "requests_max_retries": 8,
    "sleep_interval": 5,
    "waiting_queue_remove_after": 10,
    "fetch_site": "cross-site",
    # "fetch_site": "same-origin",
    "robot_date_format": "%Y-%m-%dT%H:%M:%S.%fZ",
    "active_home": "",
    "pending_home": "",
    "inactive_home": "",
    "paused_home": "",
    "coordinators_home": "",
    "lock_home": "",
    "gnupg_home": "",
    "log_home": "",
    "waiting_queue_file": "",
    "config_file": "",
    "message_notification_command": "",
    "lightning_node_command": "",
}



#!/usr/bin/env python3

"""setup.py"""

import setuptools


setuptools.setup(
    name="roboauto",
    version="0.3.1",
    description="robosats cli",
    license="ISC",
    author="jerryfletcher21",
    author_email="jerryfletcher@cock.email",
    packages=["roboauto"],
    scripts=["bin/roboauto"]
)
